# -*- coding: utf-8 -*-
"""
Created on Wed Mar  8 15:42:42 2023

@author: Songjian Lu
"""
import time

from __ReDeconv_N import *

# Note: How to run this script. -------------------------------------------
# 1. create a folder, such as "ReDeconv".
# 2. In "ReDeconv", create folders "demo_data_4_normalization" and "Results_4_normalization".
# 3. Copy "ReDeconv_Normalization.py" into the folder "ReDeconv".
# 4. Copy demo data "Demo_SN160K_5sp_meta.tsv" and "Demo_SN160K_5sp_scRNAseq_2.tsv" into the folder "demo_data_4_normalization".
# 5. run "ReDeconv_Normalization.py"


print('********************************************************\n')
print('  1 -- Compute gene expresion means for all cell types')
print('  2 -- Check the cell type classification quality for the scRNA-seq data (heatmap-plot)')
print('  3 -- Check the cell type classification quality for the scRNA-seq data (point-plot)')
print('  4 -- Do scRNA-seq data normalization for cell type deconvolution')
print('\n*********************************************************\n')
choice = int(input("Input your choice: "))

stTime = time.mktime(time.gmtime())

# ------ Input file name
fn_meta = './demo_data_4_normalization/Demo_SN160K_5sp_meta.tsv'
fn_exp = './demo_data_4_normalization/Demo_SN160K_5sp_scRNAseq_2.tsv'

# ------ Output file names for cell type count and mean information
fn_ctyp_mean = './Results_4_normalization/Ctype_size_means.tsv'
fn_ctyp_count = './Results_4_normalization/Ctype_cell_counts.tsv'
fn_cell_transcriptome_size = './Results_4_normalization/Cell_trans_sizes.tsv'

if choice == 1:
    # Equivalent to Step-1 of GUI version for normalization !!!
    # Compute gene expression means for all cell types

    '''
   In any sample, if the cell count of a cell type is less than "L_cell_count_Low_bound", 
   then the transcriptome size of the cell type in this sample would be set to "nan",
   which would not be used in the normalization. Default low bound is 10.
   The expresson profiles of those cells can still be normalized by using transcriptome size mean
   information of other cell types.
   '''
    # Input: fn_meta, fn_exp
    # Parameter: L_cell_count_Low_bound
    # Output: fn_ctyp_mean, fn_ctyp_count, fn_cell_transcriptome_size
    L_cell_count_Low_bound = 10

    # Check if scRNA-seq and meta data are matched.
    L_status_data = check_meta_and_scRNAseq_data(fn_meta, fn_exp)

    if (L_status_data > 0):
        get_sample_cell_type_exp_mean_and_cell_count(fn_meta, fn_exp, fn_ctyp_mean, fn_ctyp_count,
                                                     fn_cell_transcriptome_size, L_cell_count_Low_bound)

if choice == 2:
    # Equivalent to Step-2 of GUI version for normalization (draw heatmap) !!!
    # Check if expression means of all cell types in any two sample have strong linear relation.
    # If the linear relation is not strong, it is likely that cells in each cell type are not purefied enough

    # ------ output file names
    fn_heatmap = './Results_4_normalization/Heatmap_plot.png'
    fn_heatmap_matrix = './Results_4_normalization/Heatmap_plot_correlation_matrix.csv'

    # Draw heatmap of Pearson-correlation coeficients for all sample pairs
    # Input: fn_ctyp_mean
    # Output: fn_heatmap, fn_heatmap_matrix
    draw_heatmap_Pearson_all(fn_ctyp_mean, fn_heatmap, fn_heatmap_matrix)

if choice == 3:
    # Equivalent to Step-2 of GUI version for normalization (draw point-plot) !!!
    # Check if expression means of all cell types in any two sample have strong linear relation.
    # If the linear relation is not strong, it is likely that cells in each cell type are not purefied enough

    # ------ output file names
    fn_extra_info = './Results_4_normalization/Extra_information.txt'
    fn_point = './Results_4_normalization/Points_plot.png'

    # ----
    L_figureNo_eachRow = 2
    L_baseline = 0
    L_Pearson_LB = 0.95

    # This function can get the baseline sample such that we can merge the most number of samples together.
    # L_baseline = get_sample_baseline(fn_ctyp_mean, L_Pearson_LB)

    # Input: fn_ctyp_count, fn_ctyp_mean
    # Parameter: L_Pearson_LB
    # Output: fn_extra_info
    get_sample_cell_type_information_top_Pearson_2(fn_ctyp_count, fn_ctyp_mean, fn_extra_info, L_Pearson_LB)

    # Input: fn_ctyp_mean
    # Parameters: L_baseline, L_Pearson_LB, L_figureNo_eachRow, L_fit_line_withShift, L_fit_line_noShift
    # Output: fn_point

    L_fit_line_withShift = 1  # 1--draw the fit line; 0--do not draw the fit line.
    L_fit_line_noShift = 1  # 1--draw the fit line; 0--do not draw the fit line.
    draw_cell_type_size_mean_point_plot(fn_ctyp_mean, fn_point, L_baseline, L_Pearson_LB, L_figureNo_eachRow,
                                        L_fit_line_withShift, L_fit_line_noShift)

if choice == 4:
    # Equivalent to Step-3 of GUI version for normalization !!!
    # Choose cells from some samples and perform normalization

    # File name of meta information for chosen cells. This file will not be created if all cells are chosen.
    fn_meta_2 = './Results_4_normalization/Meta_data_new.tsv'

    # File name for scRNA-seq data submatrix for chosen cells. This file will not be created if all cells are chosen.
    fn_exp_2 = './Results_4_normalization/scRNA_seq_temp_file.tsv'

    # File name for normalized scRNA-seq data; no shift.
    fn_exp_3 = './Results_4_normalization/scRNA_seq_new_noShift.tsv'

    # File name for normalized scRNA-seq data; with shift
    fn_exp_4 = './Results_4_normalization/scRNA_seq_new_withShift.tsv'

    L_Pearson_LB = 0.95
    L_baseline = 0

    # This function can get the baseline sample such that we can merge the most number of samples together.
    # L_baseline = get_sample_baseline(fn_ctyp_mean, L_Pearson_LB)
    # print('--Base line:', L_baseline)

    # Input: fn_exp, fn_meta, fn_ctyp_mean, fn_cell_transcriptome_size
    # Parameters: L_baseline, L_Pearson_LB, L_normalization_withShift
    # Output: fn_meta_2, fn_exp_2, fn_exp_3/fn_exp_4

    L_normalization_withShift = 'N'
    if not L_normalization_withShift == 'Y':
        # Doing normalizaton: the shift is not allowed.
        print('--- Doing normalizaitn with shift not allowed!!! ---')
        get_cell_subset_scRNA_seq_data_normalization_no_shift(fn_exp, fn_meta, fn_ctyp_mean, fn_cell_transcriptome_size,
                                                              fn_exp_2, fn_meta_2, fn_exp_3, L_baseline, L_Pearson_LB)
    else:
        # Doing normalization: the shift is allowed.
        print('--- Doing normalizaitn with shift allowed ---')
        get_cell_subset_scRNA_seq_data_normalization(fn_exp, fn_meta, fn_ctyp_mean, fn_cell_transcriptome_size,
                                                     fn_exp_2, fn_meta_2, fn_exp_4, L_baseline, L_Pearson_LB)

endTime = time.mktime(time.gmtime())
print('\n\nTotal time =', ((endTime - stTime) / 60), 'minutes')





# -*- coding: utf-8 -*-
"""
Created on Fri Jan 27 14:27:43 2023

@author: Songjian Lu
"""

import time

from __ReDeconv_P import *

# Note: How to run this script. -------------------------------------------
# 1. create a folder, such as "ReDeconv".
# 2. In "ReDeconv", create folders "demo_data_4_deconvolution" and "Results_4_deconvolution".
# 3. Copy "ReDeconv_Percentage.py" into the folder "ReDeconv".
# 4. Copy demo data "References_Meta_data_subset.tsv", "References_scRNA_seq_Nor.tsv", and "Synthetic_Bulk_RNA_seq_Equal_Fraction_TPM.tsv" into the folder "demo_data_4_deconvolution".
# 5. run "ReDeconv_Percentage.py"


print('********************************************************\n')
print('  1 -- Find initial signature genes (t-test)')
print('  2 -- Compute mean and std of top signature genes')
print('  3 -- Do cell type deconvolution')
print('\n*********************************************************\n')
choice = int(input("Input your choice: "))

stTime = time.mktime(time.gmtime())

# ------ Input and output file name
fn_meta = './demo_data_4_deconvolution/References_Meta_data_subset.tsv'
fn_exp = './demo_data_4_deconvolution/References_scRNA_seq_Nor.tsv'

fn_ini_sig = './Results_4_deconvolution/Initial_sig_t_test_fd2.0_corr.tsv'
fn_mean_std = './Results_4_deconvolution/Signature_mean_std_fd2.0.tsv'
fn_heatmap = './Results_4_deconvolution/Heatmap_signature_gene_matrix.png'  # new file name
fn_extra_info = './Results_4_deconvolution/Signature_genes_extra_information.txt'  # new file name

fn_bulk_RNAseq_raw = './demo_data_4_deconvolution/Synthetic_Bulk_RNA_seq_Equal_Fraction_TPM.tsv'
fn_percentage_save = './Results_4_deconvolution/ReDeconv_results.tsv'

if choice == 1:
    # Equivalent to Step-1 of GUI version for deconvolution !!!
    # Use t_test (pairwise, cell types) to find initial signature genes

    # Input: fn_meta, fn_exp
    # Parameter: L_max_pv, L_min_fold_change
    # Output: fn_ini_sig

    L_max_pv = 0.05
    L_min_fold_change = 2.0
    L_CellType_CellNo_LB = 30  # new parameter
    L_NoSep_sampleNo_UB = 2  # new parameter

    # Check if scRNA-seq and meta data are matched.
    L_status_data = check_meta_and_scRNAseq_data(fn_meta, fn_exp)

    if L_status_data > 0:
        get_initial_Signature_Candidates(fn_meta, fn_exp, fn_ini_sig, L_max_pv, L_min_fold_change, L_CellType_CellNo_LB,
                                         L_NoSep_sampleNo_UB)

if choice == 2:
    # Equivalent to Step-2 of GUI version for deconvolution !!!
    # Get means and std for top signature genes chosen

    # Input: fn_meta, fn_exp, fn_ini_sig
    # Parameter: L_topNo
    # Output: fn_mean_std, fn_heatmap, fn_extra_info

    L_topNo = 133  # Upbound for number of signature genes for each cell type
    Get_signature_gene_matrix(fn_exp, fn_meta, fn_ini_sig, fn_mean_std, L_topNo, fn_heatmap, fn_extra_info)

if choice == 3:
    # Equivalent to Step-1 of GUI version for deconvolution !!!
    # Find percentages of cell types in mixture samples

    # Input: fn_mean_std, fn_bulk_RNAseq_raw
    # Parameter: NA
    # Output: fn_percentage_save

    ReDeconv(fn_mean_std, fn_bulk_RNAseq_raw, fn_percentage_save)

endTime = time.mktime(time.gmtime())
print('\n\nTotal time =', ((endTime - stTime) / 60), 'minutes')


# -*- coding: utf-8 -*-
"""
Created on Wed Mar  8 15:30:19 2023

@author: Songjian Lu

---------- Modified on Sep. 13, 2024
Add two line in point-plot: "green"--linear regression with shift; "red"--linear regression without shift.
Add the choice of no-shift for normalization
"""
import numpy, math, shutil
from numpy.linalg import inv
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt 
from matplotlib.figure import Figure



p_st_using = 'Y'

testShow ='N'

def read_meta_info_sample_ctype(p_file, p_cid_colNo = 0, p_sp_colNo=2, p_type_colNo = 1, p_sp = '\t'):
   L_rets = {}; L_ctype = {}
   ff = open(p_file, 'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln0 = ln.split('\n')
         ln1 = ln0[0].split(p_sp)
         ctype = ln1[p_type_colNo]
         cid = ln1[p_cid_colNo]
         spID = ln1[p_sp_colNo]

         L_rets[cid] = [spID, ctype]
         if not ctype in L_ctype:
            L_ctype[ctype] = 0
         L_ctype[ctype] += 1

   ff.close()
   
   if (p_st_using.upper())[0] == 'Y':
      print('--- Counts of different cell types ---')
      print('  cell Type name\tcell number')
   for tp in L_ctype:
      print('  ',tp, '\t\t', L_ctype[tp])

         
   return L_rets

def check_scRNA(p_file,p_st = 0):
   #p_st = 0:org count, 1:log count
   L_rets = {}; L_min = 10000.0; L_max = -1000.0
   L_rets2 = {}
   L_colNo = 0; L_rowNo = 0
   L_title = []
   ff = open(p_file,'r')
   LineNo = 0
   for ln in ff:
      LineNo += 1
   ff.close()
   if (p_st_using.upper())[0] == 'Y':
      print('---- Total row#:', LineNo)
   ff = open(p_file,'r')
   count = 0
   for ln in ff:
       count += 1
       ln1 = ln.split('\n')
       ln2 = ln1[0].split('\t')
       if count % 100 == 0:
          if (p_st_using.upper())[0] == 'Y':
             print(' --- Processing row:', count,'/',LineNo)
       if count<3:
          if (p_st_using.upper())[0] == 'Y':
             print(ln2[0:5],ln2[-5:],len(ln2))
         
         #print(ln)
       if count == 1:
         L_colNo = len(ln2)
         L_title = ln2
         for idx in range(1,len(ln2)):
             L_rets[idx] = 0.0
             L_rets2[idx] = 0
       else:
         for idx in range(1,len(ln2)):
             va = float(ln2[idx])
             va2 = va
             if p_st>0:
               va2 = 2**va
             if va>L_max:
               L_max = va
             if va<L_min:
               L_min = va
             L_rets[idx] += va2
             if va>0:
               L_rets2[idx] += 1
   ff.close()
   L_rowNo = count
   if (p_st_using.upper())[0] == 'Y':
      print('Row#:',L_rowNo, ' Col#:',L_colNo)
      print('Min:',L_min, ' Max:',L_max)
   for idx in range(1,10):
      if (p_st_using.upper())[0] == 'Y':
         print(idx, L_rets[idx])
   return L_rets, L_rets2, L_title


def check_counts_in_sample_type(p_exp_count, p_cid, p_cid_sp_tp):
   L_rets = {}; L_ctype_all = {}; L_cell_tsize = {}
      
   for idx in p_exp_count:
      cid = p_cid[idx]
      t_exp_v = p_exp_count[idx]
      L_cell_tsize[cid] = t_exp_v
      
      cell_sample = p_cid_sp_tp[cid][0]
      cell_type = p_cid_sp_tp[cid][1]
      L_ctype_all[cell_type] = 1
      if not cell_sample in L_rets:
         L_rets[cell_sample] = {}
      if not cell_type in L_rets[cell_sample]:
         L_rets[cell_sample][cell_type] = []
      L_rets[cell_sample][cell_type].append(t_exp_v)

   return L_rets, L_ctype_all, L_cell_tsize


def get_sample_cell_type_exp_mean(p_sp_tp_count, p_fn_out, p_fn_out_2,p_ctypeAll, p_cell_count_LB):
   fout = open(p_fn_out,'w')
   L_tp_list = []
   for tp in p_ctypeAll:
      L_tp_list.append(tp)
   L_tp_list.sort()
   L_sp_list = []
   for sp in p_sp_tp_count:
      L_sp_list.append(sp)
   L_sp_list.sort()
   output = 'sample/cell_type\t' +'\t'.join(L_tp_list)+'\n'
   fout.write(output)
   for sp in L_sp_list:
      output = sp
      for tp in L_tp_list:
         if tp in p_sp_tp_count[sp]:
            t_values = p_sp_tp_count[sp][tp]
            if len(t_values)>=p_cell_count_LB:
               t_mean = numpy.mean(t_values)
               output += '\t'+str(t_mean)
            else:
               output += '\tnan'
         else:
            output += '\tnan'
      output += '\n'
      fout.write(output)
   fout.close()

   fout = open(p_fn_out_2,'w')
   output = 'sample/cell_type\t' +'\t'.join(L_tp_list)+'\n'
   fout.write(output)
   for sp in L_sp_list:
      output = sp
      for tp in L_tp_list:
         if tp in p_sp_tp_count[sp]:
            t_values_No = len(p_sp_tp_count[sp][tp])
            output += '\t'+str(t_values_No)
         else:
            output += '\t0'
      output += '\n'
      fout.write(output)  
   fout.close()


def check_meta_and_scRNAseq_data(p_meta, p_exp):
   L_rets = 1
   
   ff = open(p_exp, 'r')
   L_title = []
   count = 0
   for ln in ff:
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      count += 1
      if count == 1:
         L_title = ln2
      else:
         if len(ln2)==len(L_title):
            break
         L_rets = -1
         output = '\n !!! In the expression file, the numbers of columns in the first and second rows are not equal ---'
         output += '\n   --- Number of columns in the fist row: ' + str(len(L_title))
         output += '\n   --- Number of columns in the second row: ' + str(len(ln2))
         output += '\n !!! Please check the input data files before run the program --- '
                  

         print(output)
         break
   ff.close()
   if L_rets<0:
      return L_rets
   
   L_cell_all = {}; L_cell_list = []; L_sample_all = {}
   ff = open(p_meta, 'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln1 = ln.split('\t')
         L_cell_all[ln1[0]] = 1
         L_cell_list.append(ln1[0])
         if len(ln1)>2:
             L_sample_all[ln1[2]] = 1            
   ff.close()
   
   if len(L_sample_all)<2:
       L_rets = -1
       output = ''
       if len(L_sample_all)==1:
         output += '\n !!! The scRNA-seq data only has one sample/subject ---'
         output += '\n !!! There is no need for the scRNA-seq data normalization ---'
       else:
         output += '\n !!! The meta data does not have the third column for the sample/subject information ---'
         output += '\n !!! There is not enough information for performing the scRNA-seq data normalization --- '

       output += '\n !!! Please check the input data files before run the program --- '
               
       print(output)
       return L_rets
   
   if len(L_cell_list)+1 == len(L_title):
      t_matched = 0
      for idx in range(1, len(L_title)):
         if L_title[idx] in L_cell_all:
            t_matched += 1
      if t_matched<len(L_cell_all):
         L_rets = -1
         output = '\n !!! Some cell IDs in the meta and expression data files are not matched ---'
         output += '\n   --- Number of cells in the expression data file: ' + str(len(L_title)-1)
         output += '\n   --- Number of cells in the meta data file: ' + str(len(L_cell_all))
         output += '\n   --- Number of cells mathced in both files: ' + str(t_matched)
         output += '\n !!! Please check the input data files before run the program --- '
                  

         print(output)         
   else:
         L_rets = -1
         output = '\n !!! The numbers of cells in the meta and expression data files are not equal ---'
         output += '\n   --- Number of cells in the expression data file: ' + str(len(L_title)-1)
         output += '\n   --- Number of cells in the meta data file: ' + str(len(L_cell_list))
         output += '\n !!! Please check the input data files before run the program --- '
                  
         print(output)
   return L_rets

def get_sample_cell_type_exp_mean_and_cell_count(p_meta, p_exp, p_out_mean, p_out_count, p_out_cell_size, p_cell_count_LB = 10):
   L_cell_types =  read_meta_info_sample_ctype(p_meta)
   L_cell_read_counts, L_cell_expressed_gene_counts, L_title = check_scRNA(p_exp)
   L_count_cell_sample_types, L_all_types, L_cell_transcriptome_size = check_counts_in_sample_type(L_cell_read_counts,L_title,L_cell_types)
   get_sample_cell_type_exp_mean(L_count_cell_sample_types, p_out_mean, p_out_count, L_all_types, p_cell_count_LB)
   
   fout = open(p_out_cell_size, 'w')
   for cid in L_cell_transcriptome_size:
      output = cid+'\t'+str(L_cell_transcriptome_size[cid])+'\n'
      fout.write(output)
   fout.close()

def read_sp_tp_mean(p_file):
   L_rets = {}; L_gid_idx = {}
   ff = open(p_file,'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln1 = ln.split('\n')
         ln2 = ln1[0].split('\t')
         gid = ln2[0]
         L_rets[gid] = []
         L_gid_idx[count-2] = gid
         for idx in  range(1,len(ln2)):
            item = ln2[idx]
            if item == 'nan':
               L_rets[gid].append(-10000)
            else:
               L_rets[gid].append(float(item))
   ff.close()
   # print('----------------------')
   # print('Index\t\tSample_name')
   # for idx in L_gid_idx:
   #    print(' ',idx,'\t\t',L_gid_idx[idx])

   return L_rets


def get_means_two_samples(p_mean_all, p_sp1, p_sp2):
   L_sp_list_all = []
   for sp in p_mean_all:
      L_sp_list_all.append(sp)
   L_sp_list_all.sort()
   

   L_sp1 = L_sp_list_all[p_sp1]
   L_sp2 = L_sp_list_all[p_sp2]

   if (p_st_using.upper())[0] == 'Y':   
      print(' --- Draw grap for samples:', L_sp1, 'and', L_sp2)

   L_data_1 = p_mean_all[L_sp1]
   L_data_2 = p_mean_all[L_sp2]

   Y_list = []
   X_list = []
   for idx in range(len(L_data_1)):
       v1 = L_data_1[idx]
       v2 = L_data_2[idx]
       if v1>=0 and v2>=0:
         X_list.append(v1)
         Y_list.append(v2)
   return X_list, Y_list

def get_LR(p_sp_tp_mean, p_chosen_s, p_fix_idx = 0):
   L_rets = {}
   L_fixed_sp = ''
   count = -1
   L_sp_list = []
   for sp in p_sp_tp_mean:
      L_sp_list.append(sp)
   L_sp_list.sort()
   for sp in L_sp_list:
      count += 1
      if count == p_fix_idx:
         L_fixed_sp = sp
         break
   if (p_st_using.upper())[0] == 'Y':
      print('---Fixed sp:', L_fixed_sp)
   
   L_data_1 = p_sp_tp_mean[L_fixed_sp]
   for sp in p_chosen_s:
      L_data_2 = p_sp_tp_mean[sp]
      

      Y_list = []
      X_list = []
      X_tp = []
      for idx in range(len(L_data_1)):
          v1 = L_data_1[idx]
          v2 = L_data_2[idx]
          if v1>0 and v2>0:
            X_list.append([v1,1])
            Y_list.append(v2)
            X_tp.append(v1)
            
      # print('\n---+++')
      # print(L_fixed_sp, (X_tp))
      # print(sp, (Y_list))      
      
      x_train = numpy.array(X_list, dtype=numpy.float32)
      y_train = numpy.array(Y_list, dtype=numpy.float32)
   
      x_train_t = x_train.transpose()
      x_train_mul = numpy.dot(x_train_t,x_train)   
      x_train_inv = inv(x_train_mul)   
      x_result_ma = numpy.dot(x_train_inv, x_train_t)   
      L_para = numpy.dot(x_result_ma, y_train)
      
      L_rets[sp] = [L_para[0],L_para[1]]
      # print('a=',L_para[0], 'b=', L_para[1])
   return L_rets

def scRNA_seq_normalization(p_exp, p_exp_out, p_meta, p_sp_ctp_tra_size_mean, p_cell_size, p_chosen_samples, p_fixed_sample = 0):
   L_sp_tp_mean = read_sp_tp_mean(p_sp_ctp_tra_size_mean)
   L_Para_LR = get_LR(L_sp_tp_mean, p_chosen_samples, p_fixed_sample)
    
   L_v_max = -100; L_a_min = 100
   count = 0
   for sp in L_Para_LR:
     # if (p_st_using.upper())[0] == 'Y':
     #    print(sp,'\ta =',L_Para_LR[sp][0],'\tb =',L_Para_LR[sp][1],'\tratio =',L_Para_LR[sp][1]/L_Para_LR[sp][0])
     if L_Para_LR[sp][0] < L_a_min:
         L_a_min = L_Para_LR[sp][0]
     count += 1
     if count == 1:
         L_v_max = L_Para_LR[sp][1]/L_Para_LR[sp][0]
     else:
         t_ratio = L_Para_LR[sp][1]/L_Para_LR[sp][0]
         if L_v_max<t_ratio:
           L_v_max = t_ratio
     #print('--max ratio: ',L_v_max)
   if L_a_min<=0:
     if (p_st_using.upper())[0] == 'Y':

         print('!!! As at least one of constant "a" is negative, transcriptome sizes of cell types in at least two samples have problems.\n')
         print('    It is likely the cell type classification is not purified enough. The normalization will not perform.')
     return 0
   
   ff = open(p_exp,'r')
   LineNo = 0
   for ln in ff:
     LineNo += 1
   ff.close()
   
   LineNo2 = LineNo
   #print('---- Total row#:', LineNo)  
   
   L_Para_LR_2 = {}
   for sp in L_Para_LR:
     L_Para_LR_2[sp] = [L_Para_LR[sp][0],L_Para_LR[sp][1],(L_v_max-L_Para_LR[sp][1]/L_Para_LR[sp][0])/LineNo2]
     #print('  ', sp, [L_Para_LR[sp][0],L_Para_LR[sp][1],(L_v_max-L_Para_LR[sp][1]/L_Para_LR[sp][0])/LineNo])
    
   L_cell_types =  read_meta_info_sample_ctype(p_meta)
   count = 0
   for cid in L_cell_types:
     count += 1
     # if count<20:
     #    if (p_st_using.upper())[0] == 'Y':
     #       print(cid, L_cell_types[cid], L_Para_LR_2[ L_cell_types[cid][0]])
       
   fout = open(p_exp_out, 'w')
   ff = open(p_exp,'r')
   count = 0
   L_title = []
   for ln in ff:
     count += 1
     ln1 = ln.split('\n')
     ln2 = ln1[0].split('\t')
     if count%100 == 1:
        if (p_st_using.upper())[0] == 'Y':
           print('--- Processing row:',count,'/',LineNo, 'for normlaization......')
     if count == 1:
       L_title = ln2
       fout.write(ln)
     else:
       t_list = [ln2[0]]
       for idx in range(1, len(ln2)):
           t_cid = L_title[idx]
           
           t_cell_size = p_cell_size[t_cid]
           
           t_spid = L_cell_types[t_cid][0]
           t_v_a = L_Para_LR_2[t_spid][0]
           t_shift = L_Para_LR_2[t_spid][2]
           t_v_org = float(ln2[idx])
           
           #Add shift to all genes
           t_v_new = t_v_org/t_v_a + t_shift
           t_list.append(str(t_v_new))
           
           #Add shift to genes with positive expression
           # if t_cell_size>0:
           #    t_v_new = t_v_org/t_v_a*(1 + t_shift*t_v_a/t_cell_size)              
           #    t_list.append(str(t_v_new))
           # else:
           #    t_list.append(str(t_v_org))
       output = '\t'.join(t_list)+'\n'
       fout.write(output)
   ff.close()
   fout.close()
         

def get_LR_All(p_sp_tp_mean, p_fix_idx = 0):
   L_rets = {}
   L_fixed_sp = ''
   count = -1
   L_sp_list = []
   for sp in p_sp_tp_mean:
      L_sp_list.append(sp)
   L_sp_list.sort()
   for sp in L_sp_list:
      count += 1
      if count == p_fix_idx:
         L_fixed_sp = sp
         break
   if (p_st_using.upper())[0] == 'Y':
      print('---Fixed sp:', L_fixed_sp)
   
   L_data_1 = p_sp_tp_mean[L_fixed_sp]
   for sp in p_sp_tp_mean:
      L_data_2 = p_sp_tp_mean[sp]
      

      Y_list = []
      X_list = []
      X_tp = []
      for idx in range(len(L_data_1)):
          v1 = L_data_1[idx]
          v2 = L_data_2[idx]
          if v1>0 and v2>0:
            X_list.append([v1,1])
            Y_list.append(v2)
            X_tp.append(v1)
            
      # print('\n---+++')
      # print(L_fixed_sp, (X_tp))
      # print(sp, (Y_list))      
      
      x_train = numpy.array(X_list, dtype=numpy.float32)
      y_train = numpy.array(Y_list, dtype=numpy.float32)
   
      x_train_t = x_train.transpose()
      x_train_mul = numpy.dot(x_train_t,x_train)   
      x_train_inv = inv(x_train_mul)   
      x_result_ma = numpy.dot(x_train_inv, x_train_t)   
      L_para = numpy.dot(x_result_ma, y_train)
      
      L_rets[sp] = [L_para[0],L_para[1]]
      # print('a=',L_para[0], 'b=', L_para[1])
   return L_rets

def scRNA_seq_normalization_All(p_exp, p_exp_out, p_meta, p_sp_ctp_tra_size_mean, p_cell_size, p_fixed_sample = 0):
   L_sp_tp_mean = read_sp_tp_mean(p_sp_ctp_tra_size_mean)
   L_Para_LR = get_LR_All(L_sp_tp_mean, p_fixed_sample)
    
   L_v_max = -100; L_a_min = 100
   count = 0
   for sp in L_Para_LR:
     # if (p_st_using.upper())[0] == 'Y':
     #    print(sp,'\ta =',L_Para_LR[sp][0],'\tb =',L_Para_LR[sp][1],'\tratio =',L_Para_LR[sp][1]/L_Para_LR[sp][0])
     if L_Para_LR[sp][0] < L_a_min:
         L_a_min = L_Para_LR[sp][0]
     count += 1
     if count == 1:
         L_v_max = L_Para_LR[sp][1]/L_Para_LR[sp][0]
     else:
         t_ratio = L_Para_LR[sp][1]/L_Para_LR[sp][0]
         if L_v_max<t_ratio:
           L_v_max = t_ratio
     #print('--max ratio: ',L_v_max)
   if L_a_min<=0:
     if (p_st_using.upper())[0] == 'Y':

         print('!!! As at least one of constant "a" is negative, transcriptome sizes of cell types in at least two samples have problems.\n')
         print('    It is likely the cell type classification is not purified enough. The normalization will not perform.')
     return 0
   
   ff = open(p_exp,'r')
   LineNo = 0
   for ln in ff:
     LineNo += 1
   ff.close()
   LineNo2 = LineNo
   #print('---- Total row#:', LineNo)  
   
   L_Para_LR_2 = {}
   for sp in L_Para_LR:
     L_Para_LR_2[sp] = [L_Para_LR[sp][0],L_Para_LR[sp][1],(L_v_max-L_Para_LR[sp][1]/L_Para_LR[sp][0])/LineNo2]
     #print('  ', sp, [L_Para_LR[sp][0],L_Para_LR[sp][1],(L_v_max-L_Para_LR[sp][1]/L_Para_LR[sp][0])/LineNo])
    
   L_cell_types =  read_meta_info_sample_ctype(p_meta)
   count = 0
   for cid in L_cell_types:
     count += 1
     # if count<20:
     #    if (p_st_using.upper())[0] == 'Y':
     #       print(cid, L_cell_types[cid], L_Para_LR_2[ L_cell_types[cid][0]])
       
   fout = open(p_exp_out, 'w')
   ff = open(p_exp,'r')
   count = 0
   L_title = []
   for ln in ff:
     count += 1
     ln1 = ln.split('\n')
     ln2 = ln1[0].split('\t')
     if count%100 == 1:
        if (p_st_using.upper())[0] == 'Y':
           print('--- Processing row:',count,'/',LineNo, 'for normlaization......')
     if count == 1:
       L_title = ln2
       fout.write(ln)
     else:
       t_list = [ln2[0]]
       for idx in range(1, len(ln2)):
           t_cid = L_title[idx]
           
           t_cell_size = p_cell_size[t_cid]
           
           t_spid = L_cell_types[t_cid][0]
           t_v_a = L_Para_LR_2[t_spid][0]
           t_shift = L_Para_LR_2[t_spid][2]
           t_v_org = float(ln2[idx])

           #Add shift to all genes
           t_v_new = t_v_org/t_v_a + t_shift
           t_list.append(str(t_v_new))
           
           #Add shift to genes with positive expression
           # if t_cell_size>0:
           #    t_v_new = t_v_org/t_v_a*(1 + t_shift*t_v_a/t_cell_size)
           #    t_list.append(str(t_v_new))
           # else:
           #    t_list.append(str(t_v_org))
           
       output = '\t'.join(t_list)+'\n'
       fout.write(output)
   ff.close()
   fout.close()
      
def get_means_two_samples_2(p_mean_all, p_sp1, p_sp2):

   L_data_1 = p_mean_all[p_sp1]
   L_data_2 = p_mean_all[p_sp2]

   Y_list = []
   X_list = []
   for idx in range(len(L_data_1)):
       v1 = L_data_1[idx]
       v2 = L_data_2[idx]
       if v1>0 and v2>0:
         X_list.append(v1)
         Y_list.append(v2)
   return X_list, Y_list

def get_Pearson_correlation_samples(p_sp_tp_mean, p_cor_lb=0.95):
   L_Pearson = {}
   L_sp_list = []
   L_Pearson_spNo = {}
   L_cor_lb = 0.95
   for sp in p_sp_tp_mean:
       L_sp_list.append(sp)
   L_sp_list.sort()
   for tsp1 in L_sp_list:
       L_Pearson[tsp1] = []
       L_Pearson_spNo[tsp1] = []
       for tsp2 in L_sp_list:
          L_v1, L_v2 =  get_means_two_samples_2(p_sp_tp_mean, tsp1, tsp2)
          if len(L_v1)>3:
            t_pc = numpy.corrcoef(L_v1, L_v2)[0][1]
            L_Pearson[tsp1].append(t_pc)
            if t_pc>=p_cor_lb:
               L_Pearson_spNo[tsp1].append(tsp2)
          else:
            L_Pearson[tsp1].append(0)
   # print('----------------------')
   # print('Index\t\tSample_name\tStrong_correlation_Sample#')            
   # for idx in range(len(L_sp_list)):
   #    tsp = L_sp_list[idx]
   #    print(idx,'\t',tsp,'\t',len(L_Pearson_spNo[tsp]))
   #    print('   ---', L_Pearson_spNo[tsp])
   return L_Pearson, L_Pearson_spNo


def draw_heatmap(p_Pearson):
   df_corr = pd.DataFrame(p_Pearson)
   #df.columns = L_sp_list
   L_col_name = list(df_corr.columns)
   df_corr.index = L_col_name
   
   
   sns.light_palette("blue", as_cmap=True)
   #sns.color_palette("light:b", as_cmap=True)
   ax = sns.heatmap(df_corr, annot=False)
   
   return ax, df_corr


def draw_heatmap_Pearson_all(p_fn_ctyp_mean, p_fn_heatmap, p_fn_matrix):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, 0.95)
   #print(' --- Total sample#:', len(L_sp_tp_mean))

   L_figure, L_matrix = draw_heatmap(L_Pearson)
   L_heatmap = L_figure.get_figure()
   L_heatmap.savefig(p_fn_heatmap, dpi=400)
   #plt.savefig(p_fn_heatmap, dpi=400)
   
   L_matrix.to_csv(p_fn_matrix)
   
   
def draw_point(p_sp_tp_mean, p_idx1, p_idx2):
 
   L_X, L_Y = get_means_two_samples(p_sp_tp_mean, p_idx1, p_idx2)

   #draw the expression means of all cell types in two chosen samples 
   
   print('X:', L_X)
   print('Y:', L_Y)
   plt.plot(L_X, L_Y, 'o')
   plt.show()
   return L_X, L_Y

def draw_cell_type_size_mean_two_samples(p_fn_ctyp_mean):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   
   
   #Choose any two samples to draw point graph
   t_sp_1_idx = int(input("Input the index of sample1 (0-"+str(len(L_sp_tp_mean)-1)+"): "))
   t_sp_2_idx = int(input("Input the index of sample2 (0-"+str(len(L_sp_tp_mean)-1)+"): "))
   tX, tY = draw_point(L_sp_tp_mean, t_sp_1_idx, t_sp_2_idx)

def get_sample_baseline(p_fn_ctyp_mean, L_Pearson_LB):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, L_Pearson_LB)
   L_sp_list = []
   for sp in L_samples:
      L_sp_list.append(sp)
   L_sp_list.sort()
   
   L_rets = 0
   L_lenTp = len(L_samples[L_sp_list[0]])
   for idx in range(len(L_sp_list)):
      tsp = L_sp_list[idx]
      if L_lenTp<len(L_samples[tsp]):
         L_rets = idx
         L_lenTp = len(L_samples[tsp])
   return L_rets


def get_W0_W1_W(X, Y):
   X_sum = 0.0; Y_sum = 0.0
   XY = 0.0; X_squre = 0.0
   
   t_N = len(X)
   for idx in range(len(X)):
      t_x = X[idx]; t_y = Y[idx]
      X_sum += t_x
      Y_sum += t_y
      XY += t_x*t_y
      X_squre += t_x*t_x

   
   # t_W1 = (XY-X_sum*Y_sum/t_N)/(X_squre-X_sum*X_sum/t_N)
   # t_W0 = Y_sum/t_N-t_W1*X_sum/t_N
   
   t_W1 = (t_N*XY-X_sum*Y_sum)/(t_N*X_squre-X_sum*X_sum)
   t_W0 = (Y_sum-t_W1*X_sum)/t_N
   t_W = XY/X_squre


   x_min = numpy.min(X)
   x_max = numpy.max(X)

   
   L_P0 = [x_min, x_max]
   L_P1 = [t_W1*x_min+t_W0, t_W1*x_max+t_W0]
   L_P2 = [t_W*x_min, t_W*x_max]

   #print('--W1, W0, W, Wr', t_W1, t_W0, t_W, t_Wr)
   return L_P0, L_P1, L_P2

def draw_cell_type_size_mean_point_plot(p_fn_ctyp_mean, p_fn_point, L_base, L_Pearson_LB, p_gn_row=3, p_fl_sft = 1, p_fl_noSft = 0):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, L_Pearson_LB)
   L_sp_list = []
   for sp in L_samples:
      L_sp_list.append(sp)
   L_sp_list.sort()
   L_sp_base = L_sp_list[L_base]
   L_chosen_sp = L_samples[L_sp_base]
   #print(L_sp_base, L_chosen_sp)
   
   if len(L_chosen_sp)<2:
      print('\n---- Only one sample is chosen, no point plot figure!!!! \n')
      return 1
   
   count = 0
   colNo = p_gn_row
   rowNo = math.ceil((len(L_chosen_sp)-1)/colNo)
   
   if len(L_chosen_sp)>2:
   
      fig, axs = plt.subplots(rowNo, colNo, figsize=(6*colNo,4.5*rowNo))
      for t_sp2 in L_chosen_sp:
         if not t_sp2 == L_sp_base:
            L_v1, L_v2 =  get_means_two_samples_2(L_sp_tp_mean, L_sp_base, t_sp2)
            t_P0, t_P1, t_P2 = get_W0_W1_W(L_v1, L_v2)
            
            t_c_idx = count % colNo
            t_r_idx = math.floor(count/colNo)
            # print('\n-----',L_sp_base,t_sp2, t_r_idx, t_c_idx)
            # print(L_v1)
            # print(L_v2)
            #axs[t_r_idx,t_c_idx].plot(L_v1, L_v2, 'o')
            if colNo == 1:
               axs[t_r_idx].plot(L_v1, L_v2, 'o')
               if not p_fl_sft == 0:
                  axs[t_r_idx].plot(t_P0, t_P1, '--', color = 'green', label='With_shift')
               if not p_fl_noSft == 0:
                  axs[t_r_idx].plot(t_P0, t_P2, '--', color = 'red', label='No_shift')
                  
               axs[t_r_idx].set_xlabel(L_sp_base)
               axs[t_r_idx].set_ylabel(t_sp2)
               if p_fl_sft>0 or p_fl_noSft>0:
                  axs[t_r_idx].legend(title = 'Fit line')
            else:
               if rowNo == 1:
                  axs[t_c_idx].plot(L_v1, L_v2, 'o')
                  if not p_fl_sft == 0:
                     axs[t_c_idx].plot(t_P0, t_P1, '--', color = 'green', label='With_shift') 
                  if not p_fl_noSft == 0:
                     axs[t_c_idx].plot(t_P0, t_P2, '--', color = 'red', label='No_shift')
                     
                  axs[t_c_idx].set_xlabel(L_sp_base)
                  axs[t_c_idx].set_ylabel(t_sp2)
                  if p_fl_sft>0 or p_fl_noSft>0:
                     axs[t_c_idx].legend(title = 'Fit line')
               else:
                  axs[t_r_idx,t_c_idx].plot(L_v1, L_v2, 'o')
                  if not p_fl_sft == 0:
                     axs[t_r_idx,t_c_idx].plot(t_P0, t_P1, '--', color = 'green', label='With_shift') 
                  if not p_fl_noSft == 0:
                     axs[t_r_idx,t_c_idx].plot(t_P0, t_P2, '--', color = 'red', label='No_shift')
                     
                  axs[t_r_idx,t_c_idx].set_xlabel(L_sp_base)
                  axs[t_r_idx,t_c_idx].set_ylabel(t_sp2)
                  if p_fl_sft>0 or p_fl_noSft>0:
                     axs[t_r_idx,t_c_idx].legend(title = 'Fit line')
            count += 1
   
            
      fig.savefig(p_fn_point, dpi=400)
   else:
      if len(L_chosen_sp) == 2:
            for t_sp2 in L_chosen_sp:
               if not t_sp2 == L_sp_base:
                  L_v1, L_v2 =  get_means_two_samples_2(L_sp_tp_mean, L_sp_base, t_sp2)

                  t_P0, t_P1, t_P2 = get_W0_W1_W(L_v1, L_v2)                  
                            
                  figure = Figure(figsize=(8,7.5), dpi=100)
                  a = figure.add_subplot(111)
                  a.plot(L_v1, L_v2, 'o') 
                  if not p_fl_sft == 0:
                     a.plot(t_P0, t_P1, '--', color = 'green', label='With_shift')
                  if not p_fl_noSft == 0:
                     a.plot(t_P0, t_P2, '--', color = 'red', label='No_shift')
                  a.legend(title = 'Fit line')
                  
                  figure.savefig(p_fn_point, dpi=400)
                  
                  plt.plot(L_v1, L_v2, 'o') 
                  
                  if not p_fl_sft == 0:
                     plt.plot(t_P0, t_P1, '--', color = 'blue', label='With_shift') 
                  if not p_fl_noSft == 0:
                     plt.plot(t_P0, t_P2, '--', color = 'red', label='No_shift') 
                  if p_fl_sft>0 or p_fl_noSft>0:
                     plt.legend(title = 'Fit line')
                  plt.show()
                  
      
def get_sp_ctp_cellNo(p_sp_ctp, p_samples):
   L_rets = {}
   for sp in p_samples:
      for ctp in p_sp_ctp[sp]:
         if not ctp in L_rets:
            L_rets[ctp] = 0
         L_rets[ctp] += p_sp_ctp[sp][ctp]
   L_ctp_list = []
   for ctp in L_rets:
      L_ctp_list.append(ctp)
   L_ctp_list.sort()
   output1 = ''
   output2 = ''
   for ctp in L_ctp_list:
      output1 += '\t'+ctp
      output2 += '\t'+str(L_rets[ctp])
   return output1, output2
   
   
   
def get_cellType_cellNumber(p_meta, p_samples_Pearson):
   L_sample_cellType_map = {}
   ff = open(p_meta, 'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln1 = ln.split('\n')
         ln2 = ln1[0].split('\t')
         tctp = ln2[1]
         tspid = ln2[2]
         if not tspid in L_sample_cellType_map:
            L_sample_cellType_map[tspid] = {}
         if not tctp in L_sample_cellType_map[tspid]:
            L_sample_cellType_map[tspid][tctp] = 0
         L_sample_cellType_map[tspid][tctp] += 1
   ff.close()
   L_sp_list = []
   for sp in p_samples_Pearson:
      L_sp_list.append(sp)
   L_sp_list.sort()
   
   
   for idx in range(len(L_sp_list)):
      tsp = L_sp_list[idx]
      print('\n',idx,'\t',tsp,'\t',len(p_samples_Pearson[tsp]),p_samples_Pearson[tsp])
      t_output1, t_output2 = get_sp_ctp_cellNo(L_sample_cellType_map, p_samples_Pearson[tsp])
      print(t_output1)
      print(t_output2 )
      #print(p_samples_Pearson[tsp])


def get_sample_cell_type_information_top_Pearson(p_fn_meta, p_fn_ctyp_mean, p_Pearson_LB = 0.7):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, p_Pearson_LB)
   # print(' --- Total sample#:', len(L_sp_tp_mean))

   
   get_cellType_cellNumber(p_fn_meta, L_samples)   

def get_cellType_cellNumber_2(p_count, p_samples_Pearson):
   L_sample_cellType_map = {}
   
   L_rets = {}
   L_title = []
   ff = open(p_count, 'r')
   count = 0
   for ln in ff:
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      count += 1
      if count == 1:
         L_title = ln2
      else:
         tsp = ln2[0]
         L_sample_cellType_map[tsp] = {}
         for idx in range(1, len(ln2)):
            tpid = L_title[idx]
            t_cno = int(ln2[idx])
            L_sample_cellType_map[tsp][tpid] = t_cno
   ff.close()  
   
   L_sp_list = []
   for sp in p_samples_Pearson:
      L_sp_list.append(sp)
   L_sp_list.sort()
   
   output_All = ''
   for idx in range(len(L_sp_list)):
      tsp = L_sp_list[idx]
      #print('\n',idx,'\t',tsp,'\t',len(p_samples_Pearson[tsp]),p_samples_Pearson[tsp])
      output = '\n'+str(idx) +'\t' + tsp +'\t' + str(len(p_samples_Pearson[tsp])) + '\t'+ str(p_samples_Pearson[tsp]) +'\n'
      output_All += output
      
      
      t_output1, t_output2 = get_sp_ctp_cellNo(L_sample_cellType_map, p_samples_Pearson[tsp])
      #print(t_output1)
      output = t_output1+'\n'
      output_All += output
      #print(t_output2)
      output = t_output2+'\n'
      output_All += output
      #print(p_samples_Pearson[tsp])
   return output_All

def get_sample_cell_type_information_top_Pearson_2(p_fn_count, p_fn_ctyp_mean, p_fn_extra_info, p_Pearson_LB = 0.7):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, p_Pearson_LB)
   # print(' --- Total sample#:', len(L_sp_tp_mean))

   
   L_output_All = get_cellType_cellNumber_2(p_fn_count, L_samples)
   
   fout = open(p_fn_extra_info, 'w')
   fout.write(L_output_All)
   fout.close()  
   print(L_output_All)

   
def get_chosen_samples_heatmap(p_base, p_samples_Pearson, p_threshold, p_sp_ctp_mean):
   L_sp_list = []
   L_rets = []; L_sp_ctp_mean2 = {}
   for sp in p_samples_Pearson:
      L_sp_list.append(sp)
   L_sp_list.sort()
   L_sp_base = L_sp_list[p_base]
   for idx in range(len(p_samples_Pearson[L_sp_base])):
      tva = p_samples_Pearson[L_sp_base][idx]
      if tva>=p_threshold:
         L_rets.append(L_sp_list[idx])
         
   for sp in L_rets:
      L_sp_ctp_mean2[sp] = p_sp_ctp_mean[sp]
   print('---Chose samples:', L_rets, len(L_rets))
   #print('***', p_sp_ctp_mean)
      
   L_Pearson2, L_samples2 = get_Pearson_correlation_samples(L_sp_ctp_mean2, p_threshold)
   draw_heatmap(L_Pearson2)
   
   return L_rets

def Make_small_meta_data_sample(p_fn_meta_small,p_fn_meta_small_malig, p_sp_chosen):
   fout = open(p_fn_meta_small_malig, 'w')
   output = 'Cell_ID\tCell_type\tSample_ID\n'
   fout.write(output)
   ff = open(p_fn_meta_small,'r')
   count = 0
   for ln in ff:
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      ctp = ln2[2]
      if ctp in p_sp_chosen:
         fout.write(ln)
         count += 1

   ff.close()
   fout.close()
   print('--Total chosen cell#:', count)
      
def get_sub_matrix_sp_col_2(p_fn_exp, p_fn_meta, p_fn_out):
   #The order of columnn in the submatrix is the order of row in mata data
   L_sp_chosen = {}; L_sp_order = []
   L_cellID_idx_map = {}
   ff = open(p_fn_meta,'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln1 = ln.split('\t')
         L_sp_chosen[ln1[0]] = 1
         L_sp_order.append(ln1[0])
   ff.close()
   #print('--- Total cell# info read:', count-1)
   fout = open(p_fn_out,'w')
   ff = open(p_fn_exp,'r')
   count = 0; count2 = 0
   #L_idx_in = []
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      if count%100 == 1:
         print('---Procesing row', count, ' for new matrix ......')
      if count == 1:
         output = 'Gene_sample'
         for idx in range(1,len(ln2)):
            tcid = ln2[idx]
            L_cellID_idx_map[tcid] = idx
         for tcid in L_sp_order:
            output += '\t'+tcid
            #L_idx_in.append(idx)
         output += '\n'
         fout.write(output)
      else:
         output = ln2[0]
         for tcid in L_sp_order:
            tidx = L_cellID_idx_map[tcid]
            output += '\t'+ln2[tidx]
         output += '\n'
         fout.write(output)
            
         

            
   ff.close()
   fout.close()
   #print('---- total cell#:', len(L_idx_in))   

def get_meta_and_exp_chosen_sample(p_exp, p_meta, p_exp_out, p_meta_out, p_chosen_sp):
   L_sp_dic = {}
   for sp in p_chosen_sp:
      L_sp_dic[sp] = 1
   Make_small_meta_data_sample(p_meta, p_meta_out, L_sp_dic)
   
   print('--- Making the new scRNA-seq data matrix for chosen samples ---')
   get_sub_matrix_sp_col_2(p_exp, p_meta_out, p_exp_out)  
   
      
def get_and_normalize_cell_subset(p_fn_exp, p_fn_meta, p_fn_ctyp_mean, p_fn_cell_size, p_fn_exp_2, p_fn_meta_2, p_fn_exp_3, p_L_sample_based, p_L_Pearson_LB):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, p_L_Pearson_LB)


   L_sample1 = get_chosen_samples_heatmap(p_L_sample_based, L_Pearson, p_L_Pearson_LB, L_sp_tp_mean)

   print('--- Get cell subset......')
   get_meta_and_exp_chosen_sample(p_fn_exp, p_fn_meta, p_fn_exp_2, p_fn_meta_2, L_sample1)
   
   print('--- Normalization.....')
   L_cell_size = read_cell_transcriptome_size(p_fn_cell_size)
   scRNA_seq_normalization(p_fn_exp_2, p_fn_exp_3, p_fn_meta_2, p_fn_ctyp_mean, L_cell_size, L_sample1,  p_L_sample_based)

def read_cell_transcriptome_size(p_file):
   L_rets = {}
   ff = open(p_file, 'r')
   for ln in ff:
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      cid = ln2[0]
      tsize = float(ln2[1])
      L_rets[cid] = tsize
   ff.close()
   return L_rets

def get_cell_subset_scRNA_seq_data_normalization(p_fn_exp, p_fn_meta, p_fn_ctyp_mean, p_fn_cell_size, p_fn_exp_2, p_fn_meta_2, p_fn_exp_3, p_L_sample_based, p_L_Pearson_LB):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, p_L_Pearson_LB)


   L_sample1 = get_chosen_samples_heatmap(p_L_sample_based, L_Pearson, p_L_Pearson_LB, L_sp_tp_mean)
   
   print(len(L_sp_tp_mean), len(L_samples), len(L_sample1))
   
   L_cell_size = read_cell_transcriptome_size(p_fn_cell_size)
   
   if len(L_sample1) == len(L_sp_tp_mean):
      shutil.copyfile(p_fn_meta, p_fn_meta_2)
      if len(L_sample1)>1:
         print('--- Do normalization.....')
         scRNA_seq_normalization_All(p_fn_exp, p_fn_exp_3, p_fn_meta, p_fn_ctyp_mean, L_cell_size,  p_L_sample_based)
      else:
         output = '\n!!!! Only one sample is chosen. No need for doing normalization ------'
         shutil.copyfile(p_fn_exp, p_fn_exp_3)
         print(output)
   else:
      
      print('--- Get cell subset......')
      get_meta_and_exp_chosen_sample(p_fn_exp, p_fn_meta, p_fn_exp_2, p_fn_meta_2, L_sample1)
      
      if len(L_sample1)>1:
         print('--- Do normalization.....')
         scRNA_seq_normalization(p_fn_exp_2, p_fn_exp_3, p_fn_meta_2, p_fn_ctyp_mean, L_cell_size, L_sample1,  p_L_sample_based)
      else:
         output = '\n!!!! Only one sample is chosen. No need for doing normalization ------'
         shutil.copyfile(p_fn_exp_2, p_fn_exp_3)
         print(output)



#---Normalization without shift

def read_meta_Sample_CellType_cid(p_fn_meta, p_cid=0, p_ctp_clust=1, p_sp=2):
   L_rets = {}; L_ctp_clu_all = {}; L_cid = {}
   ff = open(p_fn_meta, 'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln1 = ln.split('\n')
         ln2 = ln1[0].split('\t')
         cid = ln2[p_cid]
         ctp = ln2[p_ctp_clust]
         tsp = ln2[p_sp]
         if not tsp in L_rets:
            L_rets[tsp] = {}
         if not ctp in L_rets[tsp]:
            L_rets[tsp][ctp] = {}
         L_rets[tsp][ctp][cid] = 1
         L_ctp_clu_all[ctp] = 1
         
         L_cid[cid] = count
         
   ff.close()
   
   if testShow=='Y':
      print('\nSample cell number in each cell type ==========================')
      for tsp in L_rets:
         print(tsp, '-------------')
         for ctp in L_rets[tsp]:
            print(' ', ctp,'\t', len(L_rets[tsp][ctp]))
            
   print('====', len(L_cid), count)
            
   return L_rets, L_ctp_clu_all

def get_CellType_Sample_ftSize_CellNo(p_fn_tfsz, p_meta):
   L_rets = {}
   ff = open(p_fn_tfsz, 'r')
   count = 0
   L_title = []
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      if count == 1:
         L_title = ln2
      else:
         #ctp = ln2[0]
         tsp = ln2[0]
         for idx in range(1, len(ln2)):
            #tsp = L_title[idx]
            ctp = L_title[idx]
            tva = ln2[idx]
            if not tva == 'nan':
               tva2 = float(tva)
               tcellNO = len(p_meta[tsp][ctp])
               if not ctp in L_rets:
                  L_rets[ctp] = {}
               L_rets[ctp][tsp] = [tcellNO, tva2]
   ff.close()
   if testShow=='Y':
      for ctp in L_rets:
         print(ctp, '-----')
         for sp in L_rets[ctp]:
            print(' ', sp, L_rets[ctp][sp])
   
   return L_rets

def get_normalization_ratio_baseLine_and_anotherSample_simple_mean(p_ctp_ftSize_CellNo, p_baseLine_sp, p_sample_2):
   #w is to minimize  = (y1-wx1)^2 + (y2-wx2)^2 + ... ; adjust ratio = 1/w; linear regression with w0=0
   L_rets = -1
   L_ratio_all = []; L_size_bl = []; L_size_anotherOne = []
   if testShow=='Y':
      print('Samples:', p_baseLine_sp, p_sample_2)
   L_deno = 0.0; L_neum = 0.0
   for ctp in p_ctp_ftSize_CellNo:
      if p_baseLine_sp in p_ctp_ftSize_CellNo[ctp] and p_sample_2 in p_ctp_ftSize_CellNo[ctp]:
         tsz_bl = p_ctp_ftSize_CellNo[ctp][p_baseLine_sp][1]
         tsz_other_sp = p_ctp_ftSize_CellNo[ctp][p_sample_2][1]
         # tRatio = tsz_bl/tsz_other_sp
         # L_ratio_all.append(tRatio)
         L_deno += tsz_bl*tsz_bl
         L_neum += tsz_other_sp*tsz_bl
         L_size_bl.append(tsz_bl)
         L_size_anotherOne.append(tsz_other_sp)
         if testShow=='Y':
            print(' ', ctp, '\t', tsz_bl, '\t', tsz_other_sp)
            
   #L_rets = L_deno/L_neum #Causeing "float division by zero" if the baseline sample and any other sample do not share any cell type. Will be changed.
   #--- Modifying begins.
   L_rets = 0.0
   if L_neum > 0:
       L_rets = L_deno/L_neum
   #--- Modifying ends. 03/22/2025.
   
   L_Pearson_coef = 0
   L_size_bl.append(0.0); L_size_anotherOne.append(0.0)
   if len(L_size_bl)>=3:
      L_Pearson_coef = t_pc = numpy.corrcoef(L_size_bl, L_size_anotherOne)[0][1]
   if testShow=='Y':
      print('Ratio mean:', L_rets, ' Pearson Coef:', L_Pearson_coef)
   return L_rets

def get_normalization_ratio_sample_mean(p_ctp_ftSize_CellNo, p_sp_ctp_cid, p_base_line_sample):
   L_rets = {}
   L_sp_list = []
   for sp in p_sp_ctp_cid:
      L_sp_list.append(sp)
   L_sp_list.sort()
   
   L_baseLine_sp = L_sp_list[p_base_line_sample]
   
   for idx in range(len(L_sp_list)):
      if not idx == p_base_line_sample:
         t_sample_2 = L_sp_list[idx]
         t_ratio = get_normalization_ratio_baseLine_and_anotherSample_simple_mean(p_ctp_ftSize_CellNo, L_baseLine_sp, t_sample_2)
         L_rets[t_sample_2] = t_ratio
   L_rets[L_baseLine_sp] = 1.0 

         
   return L_rets

def get_normalization_ratio_cell_type_sp(p_ctp_ftSize_CellNo, p_sp_ratio, p_adjust_status = 0):
   #1. Compute the transcriptome size of each cell type: weight sum of size in each sample. Weight of sp1 = (sp1_cell#)/(sp1_cell# + sp2_cell#+...)
   #2. Get adjust of each cell type and each sample.
   #3. If p_adjust_status ==0, not adjust transcriptome size of the same cell type in different samples to the same value.
   L_rets = {}
   if p_adjust_status == 0:
      for ctp in p_ctp_ftSize_CellNo:
         L_rets[ctp] = {}
         for sp in p_ctp_ftSize_CellNo[ctp]:
             L_rets[ctp][sp] = p_sp_ratio[sp]
   else:
      for ctp in p_ctp_ftSize_CellNo:
         L_rets[ctp] = {}
         #print(ctp, '===========')
         tsum_fts = 0.0
         t_cNo_all = 0.0
         for sp in p_ctp_ftSize_CellNo[ctp]:
             tratio = p_sp_ratio[sp]
             #print('  ', sp, p_ctp_ftSize_CellNo[ctp][sp], tratio, tratio*p_ctp_ftSize_CellNo[ctp][sp][1])
             t_cNo = p_ctp_ftSize_CellNo[ctp][sp][0]
             tsum_fts += p_ctp_ftSize_CellNo[ctp][sp][1]*tratio*t_cNo
             t_cNo_all += t_cNo
         tsum_fts_avg = tsum_fts/t_cNo_all
         #print('  ** new avg ratio:', tsum_fts_avg)
         for sp in p_ctp_ftSize_CellNo[ctp]:
             t_fts = p_ctp_ftSize_CellNo[ctp][sp][1]
             tratio_new = tsum_fts_avg/t_fts
             #print(' --Ratio_new, Ratio_old', sp, tratio_new, p_sp_ratio[sp])
             L_rets[ctp][sp] = tratio_new
         
   return L_rets

def get_cid_normalization_ratio(p_normalization_ratio_CellType_sp, p_normalization_ratio_sp, p_sp_ctp_cid):
   #Get the ajust ratio for each cell
   L_rets = {}
   for sp in p_sp_ctp_cid:
      for ctp in p_sp_ctp_cid[sp]:
         if ctp in p_normalization_ratio_CellType_sp:
            if sp in p_normalization_ratio_CellType_sp[ctp]:
               tRatio = p_normalization_ratio_CellType_sp[ctp][sp]
               for cid in p_sp_ctp_cid[sp][ctp]:
                  L_rets[cid] = tRatio
            else:
               tRatio = p_normalization_ratio_sp[sp]
               for cid in p_sp_ctp_cid[sp][ctp]:
                   L_rets[cid] = tRatio                
         else:
            tRatio = p_normalization_ratio_sp[sp]
            for cid in p_sp_ctp_cid[sp][ctp]:
                L_rets[cid] = tRatio   
               
   return L_rets

def get_cellType_sample_CellNo_and_tSizeMean(p_fn_ctp_ftsz, p_fn_meta, p_base_line_sample = 0):
   
   L_sp_ctp_cid, L_ctp_info = read_meta_Sample_CellType_cid(p_fn_meta)
   
   
   L_ctp_ftSize_CellNo = get_CellType_Sample_ftSize_CellNo(p_fn_ctp_ftsz, L_sp_ctp_cid)
   
   L_normalization_ratio_sp = get_normalization_ratio_sample_mean(L_ctp_ftSize_CellNo, L_sp_ctp_cid, p_base_line_sample)
   
   # print('======')
   # for sp in L_normalization_ratio_sp:
   #     print(sp, L_normalization_ratio_sp[sp])
      
   L_normalization_ratio_CellType_sp = get_normalization_ratio_cell_type_sp(L_ctp_ftSize_CellNo, L_normalization_ratio_sp)

   # print('======')
   # for sp in L_normalization_ratio_sp:
   #     print(' old',sp, L_normalization_ratio_sp[sp], len(L_ctp_ftSize_CellNo))
      
   L_cid_ratio = get_cid_normalization_ratio(L_normalization_ratio_CellType_sp, L_normalization_ratio_sp, L_sp_ctp_cid)
   
   #print(' *** Cell Number:', len(L_cid_ratio))
   
   return L_cid_ratio


def do_normalization(p_fn_scRNAseq_old, p_fn_scRNAseq_CBTS, p_cid_ratio):
   L_idx_cid_ratio = {}; L_idx_list = []
   fout = open(p_fn_scRNAseq_CBTS, 'w')
   ff = open(p_fn_scRNAseq_old, 'r')
   count = 0
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      if count%100 == 1:
         print('--- Normalizing row', count)
      if count == 1:
         cell_chosen = [ln2[0]]
         for idx in range(1, len(ln2)):
            item = ln2[idx]
            if item in p_cid_ratio:
               L_idx_cid_ratio[idx] = p_cid_ratio[item]
               L_idx_list.append(idx)
               cell_chosen.append(item)
         output = '\t'.join(cell_chosen)
         fout.write(output)
         fout.write('\n')
      else:
         vas_all = [ln2[0]]
         for idx in L_idx_list:
            tva_old = float(ln2[idx])
            t_ratio = L_idx_cid_ratio[idx]
            tva_new = str(tva_old*t_ratio)
            vas_all.append(tva_new)
         output = '\t'.join(vas_all)
         fout.write(output)
         fout.write('\n')  

   ff.close()
   fout.close()  

def get_cell_subset_scRNA_seq_data_normalization_no_shift(p_fn_exp, p_fn_meta, p_fn_ctyp_mean, p_fn_cell_size, p_fn_exp_2, p_fn_meta_2, p_fn_exp_3, p_L_sample_based, p_L_Pearson_LB):
   L_sp_tp_mean = read_sp_tp_mean(p_fn_ctyp_mean)
   L_Pearson, L_samples = get_Pearson_correlation_samples(L_sp_tp_mean, p_L_Pearson_LB)


   L_sample1 = get_chosen_samples_heatmap(p_L_sample_based, L_Pearson, p_L_Pearson_LB, L_sp_tp_mean)
   
   print(len(L_sp_tp_mean), len(L_samples), len(L_sample1))
   
   L_ctp_ftsz_cNo = get_cellType_sample_CellNo_and_tSizeMean(p_fn_ctyp_mean, p_fn_meta, p_L_sample_based)
   
   # L_cell_size = read_cell_transcriptome_size(p_fn_cell_size)
   
   if len(L_sample1) == len(L_sp_tp_mean):
       shutil.copyfile(p_fn_meta, p_fn_meta_2)
       if len(L_sample1)>1:
         print('--- Do normalization.....')
         do_normalization(p_fn_exp, p_fn_exp_3,L_ctp_ftsz_cNo)
       else:
         output = '\n!!!! Only one sample is chosen. No need for doing normalization ------'
         shutil.copyfile(p_fn_exp, p_fn_exp_3)
         print(output)
   else:
      
       print('--- Get cell subset......')
       get_meta_and_exp_chosen_sample(p_fn_exp, p_fn_meta, p_fn_exp_2, p_fn_meta_2, L_sample1)
      
       if len(L_sample1)>1:
         print('--- Do normalization.....')
         do_normalization(p_fn_exp_2, p_fn_exp_3,L_ctp_ftsz_cNo)
       else:
         output = '\n!!!! Only one sample is chosen. No need for doing normalization ------'
         shutil.copyfile(p_fn_exp_2, p_fn_exp_3)
         print(output)





#--End of Normalization without shift


# -*- coding: utf-8 -*-
"""
Created on Wed Mar  8 11:45:43 2023

@author: Songjian Lu
"""
import scipy.stats as stats
import numpy, math, time, random
from numpy.linalg import inv
from scipy.optimize import nnls
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt  


p_st_development = 'N'
p_st_using = 'Y'

#============================================================================================

def read_meta_info(p_file, p_cid_colNo = 0, p_type_colNo = 1, p_sp = '\t'):
   L_rets = {}; L_rets2 = {}
   ff = open(p_file, 'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln0 = ln.split('\n')
         ln1 = ln0[0].split(p_sp)
         ctype = ln1[p_type_colNo]
         cid = ln1[p_cid_colNo]
         if not ctype in L_rets:
            L_rets[ctype] = {}
         L_rets[ctype][cid] = 1
         ctype2 = ctype.split('"')
         if len(ctype2)>1:
            ctype = ctype2[1]
         L_rets2[cid] = ctype

   ff.close()

   if (p_st_using.upper())[0] == 'Y':  
      print('--- Counts of different cell types ---')
   for tp in L_rets:
      if (p_st_using.upper())[0] == 'Y':
         #print(tp, '\t', len(L_rets[tp]))
         print(' +++',tp, '\t', len(L_rets[tp]))

   count = 0
   for item in L_rets2:
      count += 1
      if count<10:
         print(item, L_rets2[item])
         
   return L_rets, L_rets2


def read_meta_info_2(p_file, p_CellNo_LB=30, p_cid_colNo = 0, p_type_colNo = 1, p_sp = '\t'):
   L_rets = {}; L_rets2 = {}
   ff = open(p_file, 'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln0 = ln.split('\n')
         ln1 = ln0[0].split(p_sp)
         ctype = ln1[p_type_colNo]
         cid = ln1[p_cid_colNo]
         if not ctype in L_rets:
            L_rets[ctype] = {}
         L_rets[ctype][cid] = 1
         ctype2 = ctype.split('"')
         if len(ctype2)>1:
            ctype = ctype2[1]
         L_rets2[cid] = ctype

   ff.close()
   
   L_rets_new = {}
   for ctp in L_rets:
      if len(L_rets[ctp])>=p_CellNo_LB:
         L_rets_new[ctp] = L_rets[ctp]

   if (p_st_using.upper())[0] == 'Y':  
      print('--- Counts of different cell types ---')
   for tp in L_rets:
      if len(L_rets[tp])>=p_CellNo_LB:
         #print(tp, '\t', len(L_rets[tp]))
         print(' +++',tp, '\t', len(L_rets[tp]))
      else:
         print(' ---',tp, '\t', len(L_rets[tp]), '\tExcluded!')

   count = 0
   for item in L_rets2:
      count += 1
      if count<5:
         print(item, L_rets2[item])
         
   return L_rets_new, L_rets2

def get_initial_Signature_genes_new_2(p_exp, p_groups, p_out, p_pv=0.05, p_fd = 2.0, p_level = 3, p_grp_geneNo = 1000, p_sp_in = '\t', p_sp_out = '\t'):
   # A gene that is considered as a signature gene must be significant  
   # up-regulated in one group (Comparing with all other groups)
   L_rets = {}; L_rets2 = {}; L_grps = {}; L_grp_list = []
   for grp in p_groups:
      L_rets[grp] = {}
      L_rets2[grp] = {}

   ff = open(p_exp, 'r')
   count = 0
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split(p_sp_in)
      if count == 1:
         for idx in range(len(ln2)):
            item = ln2[idx]
            for grp in p_groups:
               if item in p_groups[grp]:
                  if not grp in L_grps:
                     L_grps[grp] = []
                  L_grps[grp].append(idx)
         t_total = 0
         for grp in L_grps:
            if (p_st_using.upper())[0] == 'Y':
               print('--Grp:', grp, len(L_grps[grp]))
            t_total += len(L_grps[grp])
            L_grp_list.append(grp)
         if (p_st_using.upper())[0] == 'Y':
            print('---Total sample:', t_total)

      else:
         if count % 100 == 0:
            if (p_st_using.upper())[0] == 'Y':
               print('--- Processing row: ',count, ' gene: ',ln2[0])
         gene = ln2[0]
         t_grp_exp = {}
         for grp in L_grp_list:
            t_grp_exp[grp] = []
            for idx in L_grps[grp]:
               t_grp_exp[grp].append(float(ln2[idx]))
         for grp in L_grp_list:
            t_exp1 = t_grp_exp[grp]
            t_st = 0; t_max_pv = 0; t_min_fd = math.inf; t_mid = 0; t_m2_zero = 0
            tmean = numpy.mean(t_exp1)
            tstd = numpy.std(t_exp1, axis=0)            
            for grp2 in L_grp_list:
               if not grp2==grp:
                  t_exp2 = t_grp_exp[grp2]
                  t_stat1_2, p_val1_2 = stats.ttest_ind(t_exp1, t_exp2, equal_var=False)

                  tmean2 = numpy.mean(t_exp2)

                  if p_val1_2<=p_pv:
                     if p_val1_2>t_max_pv:
                        t_max_pv = p_val1_2
                     if tmean2>0:
                        if tmean/tmean2>=p_fd:
                           t_st += 1
                        if t_min_fd>tmean/tmean2:
                           t_min_fd = tmean/tmean2
                     else:
                        if tmean>0:
                           t_st += 1
                           t_m2_zero += 1
            t_level = len(p_groups) - t_st
            #if t_st >= len(p_groups)-1:
            if t_level <= p_level:
               if not t_level in L_rets[grp]:
                  L_rets[grp][t_level] = {}
               L_rets[grp][t_level][gene] = [str(t_max_pv), str(t_min_fd), str(t_level), str(tmean), str(tstd)]

            
   ff.close()
   if (p_st_using.upper())[0] == 'Y':
      print('\n--- new signature genes:')
   L_total_genes = {}
   for grp in L_grp_list:
      if (p_st_using.upper())[0] == 'Y':
         print(grp, len(L_rets[grp]))
      for tlv in L_rets[grp]:
         if (p_st_using.upper())[0] == 'Y':
            print('   ',tlv, len(L_rets[grp][tlv]))
         # count = 0
         # for gene in L_rets[grp][tlv]:
         #    L_total_genes[gene] = 1
         #    count += 1
         #    if count<=2:
         #       print('   --',tlv,gene,L_rets[grp][tlv][gene])
         
   if (p_st_using.upper())[0] == 'Y':
      print('----------------')
      
      print('---Total signature#:', len(L_total_genes))        
   fout = open(p_out, 'w')
   for grp in L_rets:
      t_count_grp = 0
      t_lv_all = []
      for tlv in L_rets[grp]:
         t_lv_all.append(tlv)
      t_lv_all.sort()
      for tlv in t_lv_all:
         if t_count_grp<=p_grp_geneNo:
            for gene in L_rets[grp][tlv]:
               output = gene + p_sp_out + grp + p_sp_out + p_sp_out.join(L_rets[grp][tlv][gene])+'\n'
               fout.write(output)
            t_count_grp += len(L_rets[grp][tlv])
         else:
            break

   fout.close()
   return L_total_genes

def get_initial_Signature_genes_new_3(p_exp, p_groups, p_out, p_pv=0.05, p_fd = 2.0, p_level = 2, p_grp_geneNo = 1000, p_sp_in = '\t', p_sp_out = '\t'):
   # A gene that is considered as a signature gene must be significant  
   # up-regulated in one group (Comparing with all other groups)
   # expression total accross all cells should be larger than mean(all totals)-std
   
   Level_UB = min(len(p_groups) - 2, p_level)
   
   
   L_rets = {}; L_rets2 = {}; L_grps = {}; L_grp_list = []; L_exp_sum = {}
   for grp in p_groups:
      L_rets[grp] = {}
      L_rets2[grp] = {}
      for t_level in range(Level_UB+1):
         L_rets[grp][t_level] = {}
         
         
   ff = open(p_exp, 'r')
   count = 0
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split(p_sp_in)
      if count == 1:
         for idx in range(len(ln2)):
            item = ln2[idx]
            for grp in p_groups:
               if item in p_groups[grp]:
                  if not grp in L_grps:
                     L_grps[grp] = []
                  L_grps[grp].append(idx)
         t_total = 0
         for grp in L_grps:
            if (p_st_using.upper())[0] == 'Y':
               print('--Grp:', grp, len(L_grps[grp]))
            t_total += len(L_grps[grp])
            L_grp_list.append(grp)
         if (p_st_using.upper())[0] == 'Y':
            print('---Total sample:', t_total)

      else:
         if count % 100 == 0:
            if (p_st_using.upper())[0] == 'Y':
               print('--- Processing row: ',count, ' gene: ',ln2[0])
         gene = ln2[0]
         L_exp_sum[gene] = 0.0
         t_grp_exp = {}
         for grp in L_grp_list:
            t_grp_exp[grp] = []
            for idx in L_grps[grp]:
               t_va = float(ln2[idx])
               t_grp_exp[grp].append(t_va)
               L_exp_sum[gene] += t_va
         for grp in L_grp_list:
            t_exp1 = t_grp_exp[grp]
            t_st = 0; t_max_pv = 0; t_min_fd = math.inf; t_mid = 0; t_m2_zero = 0
            tmean = numpy.mean(t_exp1)
            tstd = numpy.std(t_exp1, axis=0)
            t_ctp_missed = []
            for grp2 in L_grp_list:
               if not grp2==grp:
                  t_exp2 = t_grp_exp[grp2]
                  t_stat1_2, p_val1_2 = stats.ttest_ind(t_exp1, t_exp2, equal_var=False)

                  tmean2 = numpy.mean(t_exp2)

                  if p_val1_2<=p_pv:
                     if p_val1_2>t_max_pv:
                        t_max_pv = p_val1_2
                     if tmean2>0:
                        if tmean/tmean2>=p_fd:
                           t_st += 1
                        else:
                           t_ctp_missed.append(grp2)
                        if t_min_fd>tmean/tmean2:
                           t_min_fd = tmean/tmean2
                     else:
                        if tmean>0:
                           t_st += 1
                           t_m2_zero += 1
                        else:
                           t_ctp_missed.append(grp2)
                  else:
                     t_ctp_missed.append(grp2)
                     if p_val1_2>t_max_pv:
                        t_max_pv = p_val1_2                     
                     if tmean2>0:
                        if t_min_fd>tmean/tmean2:
                           t_min_fd = tmean/tmean2                     
                     
            t_level = len(p_groups) - t_st - 1
            #if t_st >= len(p_groups)-1:
            if t_level <= Level_UB:
               if not t_level in L_rets[grp]:
                  L_rets[grp][t_level] = {}
               L_rets[grp][t_level][gene] = [str(t_max_pv), str(t_min_fd), str(t_level), str(tmean), str(tstd), '|'.join(t_ctp_missed)]

            
   ff.close()
   if (p_st_using.upper())[0] == 'Y':
      print('\n--- new signature genes:')
   L_total_genes = {}
   for grp in L_grp_list:
      if (p_st_using.upper())[0] == 'Y':
         print(grp, len(L_rets[grp]))
      tp_level_list = []
      for tlv in L_rets[grp]:
         tp_level_list.append(tlv)
      tp_level_list.sort()
      for tlv in tp_level_list:
         if (p_st_using.upper())[0] == 'Y':
            print('   ',tlv, len(L_rets[grp][tlv]))
         # count = 0
         for gene in L_rets[grp][tlv]:
             L_total_genes[gene] = 1
         #    count += 1
         #    if count<=2:
         #       print('   --',tlv,gene,L_rets[grp][tlv][gene])
   if (p_st_using.upper())[0] == 'Y':
      print('----------------')
   
   L_sum_all = []
   for gene in L_exp_sum:
      L_sum_all.append(L_exp_sum[gene])
   t_mean = numpy.mean(L_sum_all)
   t_std = numpy.std(L_sum_all, axis=0) 
   t_threshold_exp_sum = t_mean - t_std
   
   if (p_st_using.upper())[0] == 'Y':
      print('---Total signature#:', len(L_total_genes))        
   fout = open(p_out, 'w')
   for grp in L_rets:
      t_count_grp = 0
      t_lv_all = []
      for tlv in L_rets[grp]:
         t_lv_all.append(tlv)
      t_lv_all.sort()
      for tlv in t_lv_all:
         if t_count_grp<=p_grp_geneNo:
            t_count_in = 0
            for gene in L_rets[grp][tlv]:
               if L_exp_sum[gene]>= t_threshold_exp_sum:
                  output = gene + p_sp_out + grp + p_sp_out + p_sp_out.join(L_rets[grp][tlv][gene])+'\n'
                  fout.write(output)
                  t_count_in += 1
            #t_count_grp += len(L_rets[grp][tlv])
            t_count_grp += t_count_in
         else:
            break

   fout.close()
   return L_total_genes


def get_initial_Signature_genes_new_3b(p_exp, p_groups, p_out, p_pv=0.05, p_fd = 2.0, p_level = 2, p_grp_geneNo = 1000, p_sp_in = '\t', p_sp_out = '\t'):
   # A gene that is considered as a signature gene must be significant  
   # up-regulated in one group (Comparing with all other groups)
   # expression total accross all cells should be larger than mean(all totals)-std
   
   Level_UB = min(len(p_groups) - 2, p_level)
   
   
   L_rets = {}; L_rets2 = {}; L_grps = {}; L_grp_list = []; L_exp_sum = {}
   for grp in p_groups:
      L_rets[grp] = {}
      L_rets2[grp] = {}
      for t_level in range(Level_UB+1):
         L_rets[grp][t_level] = {}
         
         
   ff = open(p_exp, 'r')
   count = 0
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split(p_sp_in)
      if count == 1:
         for idx in range(len(ln2)):
            item = ln2[idx]
            for grp in p_groups:
               if item in p_groups[grp]:
                  if not grp in L_grps:
                     L_grps[grp] = []
                  L_grps[grp].append(idx)
         t_total = 0
         for grp in L_grps:
            if (p_st_using.upper())[0] == 'Y':
               print('--Grp:', grp, len(L_grps[grp]))
            t_total += len(L_grps[grp])
            L_grp_list.append(grp)
         if (p_st_using.upper())[0] == 'Y':
            print('---Total sample:', t_total)

      else:
         if count % 100 == 0:
            if (p_st_using.upper())[0] == 'Y':
               print('--- Processing row: ',count, ' gene: ',ln2[0])
         gene = ln2[0]
         L_exp_sum[gene] = 0.0
         t_grp_exp = {}
         t_test_pv = {}
         for grp in L_grp_list:
            t_grp_exp[grp] = []
            t_test_pv[grp] = {}
            for idx in L_grps[grp]:
               t_va = float(ln2[idx])
               t_grp_exp[grp].append(t_va)
               L_exp_sum[gene] += t_va
                              
         for grp in L_grp_list:
            for grp2 in L_grp_list:
               if grp2>grp:
                  t_stat1_2, p_val1_2 = stats.ttest_ind(t_grp_exp[grp], t_grp_exp[grp2], equal_var=False)
                  t_test_pv[grp][grp2] = p_val1_2
                  t_test_pv[grp2][grp] = p_val1_2
                  
         for grp in L_grp_list:
            t_exp1 = t_grp_exp[grp]
            t_st = 0; t_max_pv = 0; t_min_fd = math.inf; t_mid = 0; t_m2_zero = 0
            tmean = numpy.mean(t_exp1)
            tstd = numpy.std(t_exp1, axis=0)
            t_ctp_missed = []
            for grp2 in L_grp_list:
               if not grp2==grp:
                  t_exp2 = t_grp_exp[grp2]
                  #t_stat1_2, p_val1_2 = stats.ttest_ind(t_exp1, t_exp2, equal_var=False)
                  p_val1_2 = t_test_pv[grp][grp2]

                  tmean2 = numpy.mean(t_exp2)

                  if p_val1_2<=p_pv:
                     if p_val1_2>t_max_pv:
                        t_max_pv = p_val1_2
                     if tmean2>0:
                        if tmean/tmean2>=p_fd:
                           t_st += 1
                        else:
                           t_ctp_missed.append(grp2)
                        if t_min_fd>tmean/tmean2:
                           t_min_fd = tmean/tmean2
                     else:
                        if tmean>0:
                           t_st += 1
                           t_m2_zero += 1
                        else:
                           t_ctp_missed.append(grp2)
                  else:
                     t_ctp_missed.append(grp2)
                     if p_val1_2>t_max_pv:
                        t_max_pv = p_val1_2                     
                     if tmean2>0:
                        if t_min_fd>tmean/tmean2:
                           t_min_fd = tmean/tmean2                     
                     
            t_level = len(p_groups) - t_st - 1
            #if t_st >= len(p_groups)-1:
            if t_level <= Level_UB:
               if not t_level in L_rets[grp]:
                  L_rets[grp][t_level] = {}
               L_rets[grp][t_level][gene] = [str(t_max_pv), str(t_min_fd), str(t_level), str(tmean), str(tstd), '|'.join(t_ctp_missed)]

            
   ff.close()
   if (p_st_using.upper())[0] == 'Y':
      print('\n--- new signature genes:')
   L_total_genes = {}
   for grp in L_grp_list:
      if (p_st_using.upper())[0] == 'Y':
         print(grp, len(L_rets[grp]))
      tp_level_list = []
      for tlv in L_rets[grp]:
         tp_level_list.append(tlv)
      tp_level_list.sort()
      for tlv in tp_level_list:
         if (p_st_using.upper())[0] == 'Y':
            print('   ',tlv, len(L_rets[grp][tlv]))
         # count = 0
         for gene in L_rets[grp][tlv]:
             L_total_genes[gene] = 1
         #    count += 1
         #    if count<=2:
         #       print('   --',tlv,gene,L_rets[grp][tlv][gene])
   if (p_st_using.upper())[0] == 'Y':
      print('----------------')
   
   L_sum_all = []
   for gene in L_exp_sum:
      L_sum_all.append(L_exp_sum[gene])
   t_mean = numpy.mean(L_sum_all)
   t_std = numpy.std(L_sum_all, axis=0) 
   t_threshold_exp_sum = t_mean - t_std
   
   if (p_st_using.upper())[0] == 'Y':
      print('---Total signature#:', len(L_total_genes))        
   fout = open(p_out, 'w')
   for grp in L_rets:
      t_count_grp = 0
      t_lv_all = []
      for tlv in L_rets[grp]:
         t_lv_all.append(tlv)
      t_lv_all.sort()
      for tlv in t_lv_all:
         if t_count_grp<=p_grp_geneNo:
            t_count_in = 0
            for gene in L_rets[grp][tlv]:
               if L_exp_sum[gene]>= t_threshold_exp_sum:
                  output = gene + p_sp_out + grp + p_sp_out + p_sp_out.join(L_rets[grp][tlv][gene])+'\n'
                  fout.write(output)
                  t_count_in += 1
            #t_count_grp += len(L_rets[grp][tlv])
            t_count_grp += t_count_in
         else:
            break

   fout.close()
   return L_total_genes

def check_meta_and_scRNAseq_data(p_meta, p_exp):
   L_rets = 1
   
   ff = open(p_exp, 'r')
   L_title = []
   count = 0
   for ln in ff:
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      count += 1
      if count == 1:
         L_title = ln2
      else:
         if len(ln2)==len(L_title):
            break
         L_rets = -1
         output = '\n !!! In the expression file, the numbers of columns in the first and second rows are not equal ---'
         output += '\n   --- Number of columns in the fist row: ' + str(len(L_title))
         output += '\n   --- Number of columns in the second row: ' + str(len(ln2))
         output += '\n !!! Please check the input data files before run the program --- '
                  

         print(output)
         break
   ff.close()
   if L_rets<0:
      return L_rets
   
   L_cell_all = {}; L_cell_list = []
   ff = open(p_meta, 'r')
   count = 0
   for ln in ff:
      count += 1
      if count>1:
         ln1 = ln.split('\t')
         L_cell_all[ln1[0]] = 1
         L_cell_list.append(ln1[0])
   ff.close()
   if len(L_cell_list)+1 == len(L_title):
      t_matched = 0
      for idx in range(1, len(L_title)):
         if L_title[idx] in L_cell_all:
            t_matched += 1
      if t_matched<len(L_cell_all):
         L_rets = -1
         output = '\n !!! Some cell IDs in the meta and expression data files are not matched ---'
         output += '\n   --- Number of cells in the expression data file: ' + str(len(L_title)-1)
         output += '\n   --- Number of cells in the meta data file: ' + str(len(L_cell_all))
         output += '\n   --- Number of cells mathced in both files: ' + str(t_matched)
         output += '\n !!! Please check the input data files before run the program --- '
                  

         print(output)         
   else:
         L_rets = -1
         output = '\n !!! The numbers of cells in the meta and expression data files are not equal ---'
         output += '\n   --- Number of cells in the expression data file: ' + str(len(L_title)-1)
         output += '\n   --- Number of cells in the meta data file: ' + str(len(L_cell_list))
         output += '\n !!! Please check the input data files before run the program --- '
                  
         print(output)
   return L_rets

def get_initial_Signature_Candidates(p_meta, p_exp, p_ini_fn_out, p_max_pv, p_min_fold_change, p_cellNo_LB=30, p_no_sep_sampleNo=2):
   #read cell type information for all cells
   L_cid_colNo = 0  # Column index with the cell ID
   L_cell_type_colNo = 1  # Column index with the cell types
   L_sep = '\t'
   L_groups, L_map = read_meta_info_2(p_meta, p_cellNo_LB, L_cid_colNo, L_cell_type_colNo, L_sep)
   
   get_initial_Signature_genes_new_3b(p_exp,L_groups,p_ini_fn_out,p_max_pv,p_min_fold_change,p_no_sep_sampleNo)   


def chose_sig_genes_top_std_mean_ratio_3(p_fn, p_topNo, p_fn_out):
   #consider only one cell type when choosing signature genes
   L_rets = {}
   ff = open(p_fn, 'r')
   for ln in ff:
      ln0 = ln.split('\n')
      ln1 = ln0[0].split('\t')
      gene = ln1[0]
      grp = ln1[1]
      t_level = int(ln1[4])
      if not grp in L_rets:
         L_rets[grp] = {}
      if not t_level in L_rets[grp]:
         L_rets[grp][t_level] = []
      L_rets[grp][t_level].append(ln1)
        
   ff.close()
   # for grp in L_rets:
   #    print('---grp#:', grp, len(L_rets[grp]))
   
   L_std_mean_ratio_tp = {}
   for grp in L_rets:
      L_std_mean_ratio_tp[grp] = {}
      for tlv in L_rets[grp]:
         L_std_mean_ratio_tp[grp][tlv] = {}
         for item in L_rets[grp][tlv]:
            t_gene = item[0]
            t_pv = float(item[2])
            t_mean = float(item[5])
            t_std = float(item[6])
            t_missed_ctp = item[7]
            t_std_mean_ratio = t_std/t_mean
            if not t_std_mean_ratio in L_std_mean_ratio_tp[grp][tlv]:
               L_std_mean_ratio_tp[grp][tlv][t_std_mean_ratio] = []
            L_std_mean_ratio_tp[grp][tlv][t_std_mean_ratio].append([t_gene, t_pv, t_missed_ctp, t_mean, t_std])
   

            
   L_gene_top_ratio = {}
   L_gene_top_level = {}
   for grp in L_std_mean_ratio_tp:
       L_gene_top_ratio[grp] = []
       L_gene_top_level[grp] = {}
       t_level_all = []
       for tlv in L_std_mean_ratio_tp[grp]:
          t_level_all.append(tlv)
       t_level_all.sort()
       count = 0
       for tlv in t_level_all:
          t_ratio_all = []
          for t_ratio in L_std_mean_ratio_tp[grp][tlv]:
            t_ratio_all.append(t_ratio)
          t_ratio_all.sort()
          if count<=p_topNo:
             L_gene_top_level[grp][tlv] = 0
             for t_ratio in t_ratio_all:
               for item in L_std_mean_ratio_tp[grp][tlv][t_ratio]:
                   count += 1
                   if count<=p_topNo:
                      #if len(item[2])>0:
                      L_gene_top_ratio[grp].append([item[0], grp, str(tlv), str(count),  str(item[3]), str(item[4]), item[2]])
                      # else:
                      #    L_gene_top_ratio[grp].append([item[0], grp, str(tlv), str(count),  str(item[3]), str(item[4]), 'NA'])
                      L_gene_top_level[grp][tlv] += 1


   for grp in L_gene_top_level:
      if (p_st_using.upper())[0] == 'Y':
         print(grp,' -------')
         for tlv in L_gene_top_level[grp]:
               print('  ',tlv,L_gene_top_level[grp][tlv])
         

   

   fout = open(p_fn_out, 'w')
   output = 'Sig_Gene\tCell_type\tLevel\tRank\tMean\tStd\tCell_type_unseparated\n'
   fout.write(output)
   G_sig_genes = {}; L_cell_type_chosen = {}
   for grp in L_gene_top_ratio:
      L_cell_type_chosen[grp] = 1
      for item in L_gene_top_ratio[grp]:
         G_sig_genes[item[0]] = 1
         output = '\t'.join(item)
         fout.write(output)
         fout.write('\n')
         if len(item[6])>0:
            print(' ****', item[0], grp,'~',item[6])
   fout.close()


   if (p_st_using.upper())[0] == 'Y':     
      print('--- Total gene#:', len(G_sig_genes))
         
   return G_sig_genes, L_cell_type_chosen


def chose_sig_genes_top_std_mean_ratio_3_bak(p_fn, p_topNo = 100):
   #consider only one cell type when choosing signature genes
   L_rets = {}
   ff = open(p_fn, 'r')
   for ln in ff:
      ln0 = ln.split('\n')
      ln1 = ln0[0].split('\t')
      gene = ln1[0]
      grp = ln1[1]
      t_level = int(ln1[4])
      if not grp in L_rets:
         L_rets[grp] = {}
      if not t_level in L_rets[grp]:
         L_rets[grp][t_level] = []
      L_rets[grp][t_level].append(ln1)
        
   ff.close()
   # for grp in L_rets:
   #    print('---grp#:', grp, len(L_rets[grp]))
   
   L_std_mean_ratio_tp = {}
   for grp in L_rets:
      L_std_mean_ratio_tp[grp] = {}
      for tlv in L_rets[grp]:
         L_std_mean_ratio_tp[grp][tlv] = {}
         for item in L_rets[grp][tlv]:
            t_gene = item[0]
            t_pv = float(item[2])
            t_mean = float(item[5])
            t_std = float(item[6])
            t_missed_ctp = item[7]
            t_std_mean_ratio = t_std/t_mean
            if not t_std_mean_ratio in L_std_mean_ratio_tp[grp][tlv]:
               L_std_mean_ratio_tp[grp][tlv][t_std_mean_ratio] = []
            L_std_mean_ratio_tp[grp][tlv][t_std_mean_ratio].append([t_gene,t_pv, t_missed_ctp])
   

            
   L_gene_top_ratio = {}
   L_gene_top_level = {}
   for grp in L_std_mean_ratio_tp:
       L_gene_top_ratio[grp] = []
       L_gene_top_level[grp] = {}
       t_level_all = []
       for tlv in L_std_mean_ratio_tp[grp]:
          t_level_all.append(tlv)
       t_level_all.sort()
       count = 0
       for tlv in t_level_all:
          t_ratio_all = []
          for t_ratio in L_std_mean_ratio_tp[grp][tlv]:
            t_ratio_all.append(t_ratio)
          t_ratio_all.sort()
          if count<=p_topNo:
             L_gene_top_level[grp][tlv] = 0
             for t_ratio in t_ratio_all:
               for item in L_std_mean_ratio_tp[grp][tlv][t_ratio]:
                   count += 1
                   if count<=p_topNo:
                         L_gene_top_ratio[grp].append(item[0])
                         L_gene_top_level[grp][tlv] += 1


   for grp in L_gene_top_level:
      if (p_st_using.upper())[0] == 'Y':
         print(grp,' -------')
         for tlv in L_gene_top_level[grp]:
               print('  ',tlv,L_gene_top_level[grp][tlv])
         

   

   
   G_sig_genes = {}; L_cell_type_chosen = {}
   for grp in L_gene_top_ratio:
      L_cell_type_chosen[grp] = 1
      for gene in L_gene_top_ratio[grp]:
         G_sig_genes[gene] = 1


   if (p_st_using.upper())[0] == 'Y':     
      print('--- Total gene#:', len(G_sig_genes))
         
   return G_sig_genes, L_cell_type_chosen



def get_mean_std_sig_gene_cell_type_new2(p_exp, p_groups, p_top_genes, p_out, p_fn_heatmap, p_sp_in = '\t', p_sp_out = '\t'):
   # L_ini_genes = {}
   # fin = open(p_fn_initial_genes,'r')
   # count = 0
   # for ln in fin:
   #    count += 1
   #    if count>1:
   #       ln1 = ln.split('\t')
   #       L_ini_genes[ln1[0]] = 1
   # fin.close()
   
   L_rets = {}; L_grps = {}
   for grp in p_groups:
      L_grps[grp] = []
   ff = open(p_exp, 'r')
   count = 0
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split(p_sp_in)
      if count == 1:
         for idx in range(len(ln2)):
            item = ln2[idx]
            for grp in p_groups:
               if item in p_groups[grp]:
                  L_grps[grp].append(idx)
      else:
            gene = ln2[0]
            #if gene in L_ini_genes:
            if gene in p_top_genes:
               t_grp_exp = {}
               L_rets[gene] = {}
               for grp in p_groups:
                  t_grp_exp[grp] = []
                  for idx in L_grps[grp]:
                     t_grp_exp[grp].append(float(ln2[idx]))
               for grp in p_groups:
                  t_exp = t_grp_exp[grp]
                  tmean = numpy.mean(t_exp)
                  tstd = numpy.std(t_exp, axis=0)
                  L_rets[gene][grp] = [str(tmean), str(tstd)]
                  
   ff.close()
   
   cty_all = []
   for gene in L_rets:
      for ctp in L_rets[gene]:
         cty_all.append(ctp)
      break
   cty_all.sort()
   
   GeneList = []
   Cell_type_gene_mean = {}
   
   ff = open(p_out,'w')
   output = ''
   for ctp in cty_all:
      output += p_sp_out + ctp + '_mean' + p_sp_out +ctp+'_std'
   output += '\n'
   ff.write(output)
   for gene in L_rets:
       GeneList.append(gene)

       tlist = [gene]
       for grp in cty_all:
          if not grp in Cell_type_gene_mean:
             Cell_type_gene_mean[grp] = []
          Cell_type_gene_mean[grp].append(float(L_rets[gene][grp][0]))
          
          #print('\t',grp)
          #try:
          tlist.append(L_rets[gene][grp][0])
          tlist.append(L_rets[gene][grp][1])
          # except:
          #    print('***',gene,L_rets[gene])
       output = p_sp_out.join(tlist)
      # print(output)
       ff.write(output+'\n')

   ff.close()
   
   df_corr = pd.DataFrame(Cell_type_gene_mean)  
   df_corr.index = GeneList
   

   svm = sns.clustermap(df_corr, z_score=0, cmap="vlag", center=0)  
   svm.savefig(p_fn_heatmap, dpi=400)
   plt.close()
   plt.show()


def Get_signature_gene_matrix(p_exp, p_meta, p_ini_sig, p_mean_std_out, p_top_ub, p_fn_heatmap, p_fn_out):
   #read cell type information for all cells
   L_cid_colNo = 0  # Column index with the cell ID
   L_cell_type_colNo = 1  # Column index with the cell types
   L_sep = '\t'
   L_groups, L_map = read_meta_info(p_meta, L_cid_colNo, L_cell_type_colNo, L_sep)


   #Choose the top signature genes
   L_topNo = p_top_ub #Number of to signature genes for cell deconvolution
   L_top_sig_genes, L_Cell_types_Chosen = chose_sig_genes_top_std_mean_ratio_3(p_ini_sig, L_topNo, p_fn_out)
   
   #Get cellID for cell types chosen
   L_groups_new = {}
   for ctp in L_Cell_types_Chosen:
      L_groups_new[ctp] = L_groups[ctp]
      #print('---ctp', ctp, len(L_groups[ctp]))
   
   
   #get mean, std using raw count scRNA-seq file
   get_mean_std_sig_gene_cell_type_new2(p_exp,L_groups_new,L_top_sig_genes,p_mean_std_out, p_fn_heatmap)


def Get_signature_gene_matrix_GeneSet(p_exp, p_meta, p_mean_std_out, p_choose_genes, p_ini_sig, p_fn_heatmap):
   #read cell type information for all cells
   L_cid_colNo = 0  # Column index with the cell ID
   L_cell_type_colNo = 1  # Column index with the cell types
   L_sep = '\t'
   L_groups, L_map = read_meta_info(p_meta, L_cid_colNo, L_cell_type_colNo, L_sep)


   #Choose the top signature genes
   #Choose the top signature genes
   L_topNo = 10 #Number of to signature genes for cell deconvolution
   L_top_sig_genes, L_Cell_types_Chosen = chose_sig_genes_top_std_mean_ratio_3(p_ini_sig, L_topNo)
   
   #Get cellID for cell types chosen
   L_groups_new = {}
   for ctp in L_Cell_types_Chosen:
      L_groups_new[ctp] = L_groups[ctp]
      
   #L_topNo = p_top_ub #Number of to signature genes for cell deconvolution
   L_top_sig_genes = {} # chose_sig_genes_top_std_mean_ratio_3(p_ini_sig, L_topNo)
   for gene in p_choose_genes:
      L_top_sig_genes[gene] = 1
   
   
   #get mean, std using raw count scRNA-seq file
   get_mean_std_sig_gene_cell_type_new2(p_exp,L_groups_new,L_top_sig_genes,p_mean_std_out, p_fn_heatmap)
   


#============================================================================================

def read_sig_gene_mean_std(p_file):
   L_rets = {}
   ctype =  []
   ff = open(p_file, 'r')
   count = 0
   for ln in ff:
      ln1 = ln.split('\n')
      ln2 = ln1[0].split('\t')
      count += 1
      if count == 1:
         for idx in range(len(ln2)):
            if idx%2 == 1:
               t_tp = ln2[idx].split('_')
               t_tp2 = t_tp[0:-1]
               ctype.append('_'.join(t_tp2))

         #print(ctype, len(ctype))
      else:
         gene = ln2[0]
         #if gene in p_genes:
         L_rets[gene] = {}
         for idx in range(len(ctype)):
             t_tp = ctype[idx]
             tmean = ln2[2*idx+1]
             tstd = ln2[2*idx+2]
             L_rets[gene][t_tp] = [float(tmean),float(tstd)]
   ff.close()
   # for gene in L_rets:
   #    print(gene, L_rets[gene])
   
   if (p_st_using.upper())[0] == 'Y':
      print('---- Gene# in the signature matris: ', len(L_rets), ' ---Cell type#:', len(ctype))
      print('  --Cell type names:',ctype)
   return L_rets, ctype

   
def read_bulk_RNA_seq_row_for_sample(p_spInfo, p_gene_in_matrix, p_sp = '\t'):
   L_rets = {}; L_mean_std_new = {}
   ff = open(p_spInfo, 'r')
   L_title  = []
   count = 0
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split(p_sp)
      if count == 1:
         L_title = ln2
      else:
         sid = ln2[0]
         L_rets[sid] = {}
         for idx in range(1, len(ln2)):
            gene = L_title[idx]
            if gene in p_gene_in_matrix:
               tva = float(ln2[idx])
               L_rets[sid][gene] = tva
               L_mean_std_new[gene] = p_gene_in_matrix[gene]
   ff.close()
   if (p_st_using.upper())[0] == 'Y':
      print('---- Sample#:', len(L_rets))
      print('---- Number of genes in the matrix: ', len(p_gene_in_matrix), ' ---Common gene number: ', len(L_mean_std_new))
   return L_rets, L_mean_std_new

def read_bulk_RNA_seq_column_for_sample(p_spInfo, p_gene_in_matrix, p_sp = '\t'):
   L_rets = {}; L_mean_std_new = {}
   ff = open(p_spInfo, 'r')
   L_title  = []
   count = 0
   for ln in ff:
      count += 1
      ln1 = ln.split('\n')
      ln2 = ln1[0].split(p_sp)
      if count == 1:
         L_title = ln2
         for idx in range(1,len(ln2)):
            L_rets[ln2[idx]] = {}
      else:
         gene = ln2[0]
         if gene in p_gene_in_matrix:
            L_mean_std_new[gene] = p_gene_in_matrix[gene]
            for idx in range(1, len(ln2)):
               sid = L_title[idx]
               tva = float(ln2[idx])
               L_rets[sid][gene] = tva
   ff.close()
   if (p_st_using.upper())[0] == 'Y':
      print('---- Sample#:', len(L_rets))
      print('---- Number of genes in the matrix: ', len(p_gene_in_matrix), ' ---Common gene number: ', len(L_mean_std_new))
   return L_rets, L_mean_std_new

def get_cell_partition_Linear_Regression_Model(p_expression_RNAseq, p_mean_std):
   L_min_comb = {}

   L_grp_order = []
   for gene in p_mean_std:
      for grp in p_mean_std[gene]:
         L_grp_order.append(grp)
      break
   Y_list = []
   X_list = []
   L_gene_list = []
   for gene in p_expression_RNAseq:
      L_gene_list.append(gene)
   for gene in L_gene_list:
       Y_list.append(p_expression_RNAseq[gene])
       t_list = []
       for grp in   L_grp_order:
         t_list.append(p_mean_std[gene][grp][0])
       X_list.append(t_list)
   x_train = numpy.array(X_list, dtype=numpy.float32)
   y_train = numpy.array(Y_list, dtype=numpy.float32)

   x_train_t = x_train.transpose()
   
   
   x_train_mul = numpy.dot(x_train_t,x_train)
   
   x_train_inv = inv(x_train_mul)
   
   x_result_ma = numpy.dot(x_train_inv, x_train_t)
   
   L_para = numpy.dot(x_result_ma, y_train)
   
   # print('LR:',L_para)
   
   
   L_min_comb = {}; L_min_comb2 = {}
   t_total = 0.0
   for idx in range(len(L_grp_order)):
      gid = L_grp_order[idx]
      va = L_para[idx]
      t_total += va
      L_min_comb[gid] = va

   for idx in range(len(L_grp_order)):
      gid = L_grp_order[idx]
      va = L_para[idx]
      L_min_comb2[gid] = va/t_total*100      
   #L_min_comb = {'MS':L_para[0],'PN':L_para[1], 'CL':L_para[2], 'NOR':L_para[3]}
   #L_min_comb = {'MS':round(L_para[0]/10),'PN':round(L_para[1]/10), 'CL':round(L_para[2]/10), 'NOR':round(L_para[3]/10)}
   # print('LR',L_min_comb)
   
   t_m_list = []; t_va_list = []
   for gene in L_gene_list:
      t_mean = 0.0; t_va = 0.0
      for grp in L_min_comb:
         t_mean += L_min_comb[grp]*p_mean_std[gene][grp][0]
         t_va += L_min_comb[grp]*p_mean_std[gene][grp][1]*p_mean_std[gene][grp][1]
      t_m_list.append(t_mean)
      t_va_list.append(t_va)
   np_t_m_list = numpy.array(t_m_list, dtype=numpy.float32)
   np_t_va_list = numpy.array(t_va_list, dtype=numpy.float32)
   
   x_result_ma_square = x_result_ma*x_result_ma
   
   fr_m = numpy.dot(x_result_ma,np_t_m_list)
   fr_va = numpy.dot(x_result_ma_square,np_t_va_list)
   # print('fr_m',fr_m)
   # print('fr_va',fr_va)
   
   L_dist_fr = {}
   for idx in range(len(L_grp_order)):
      gid = L_grp_order[idx]
      va = fr_va[idx] 
      L_dist_fr[gid] = va
   
   # L_dist_fr = {'MS':[round(L_para[0]/10),fr_va[0]],'PN':[round(L_para[1]/10),fr_va[1]], 'CL':[round(L_para[2]/10),fr_va[2]], 'NOR':[round(L_para[3]/10),fr_va[3]]}
  # t_normalization = 1.0
   #L_dist_fr = {'MS':[round(L_para[0]/t_normalization),fr_va[0]/t_normalization],'PN':[round(L_para[1]/t_normalization),fr_va[1]/t_normalization], 'CL':[round(L_para[2]/t_normalization),fr_va[2]/t_normalization], 'NOR':[round(L_para[3]/t_normalization),fr_va[3]/t_normalization]}
   #L_dist_fr = {'MS':[L_para[0]/t_normalization,fr_va[0]/t_normalization],'PN':[L_para[1]/t_normalization,fr_va[1]/t_normalization], 'CL':[L_para[2]/t_normalization,fr_va[2]/t_normalization], 'NOR':[L_para[3]/t_normalization,fr_va[3]/t_normalization]}
  
   # print(L_dist_fr)
   
   return L_min_comb2

def get_cell_partition_Linear_Regression_Model_2(p_expression_RNAseq, p_mean_std, p_std_gene_old):
   #!!! NOT take absolute fraction value when computing variance
   #Only consider positive fractions when computing variance
   L_min_comb = {}

   L_grp_order = []
   for gene in p_mean_std:
      for grp in p_mean_std[gene]:
         L_grp_order.append(grp)
      break
   Y_list = []
   X_list = []
   L_gene_list = []
   for gene in p_expression_RNAseq:
      L_gene_list.append(gene)
   for gene in L_gene_list:
       t_gene_std = p_std_gene_old[gene] + 0.00000001
       Y_list.append(p_expression_RNAseq[gene]/t_gene_std)
       t_list = []
       for grp in   L_grp_order:
         t_list.append(p_mean_std[gene][grp][0]/t_gene_std)
       X_list.append(t_list)
   x_train = numpy.array(X_list, dtype=numpy.float32)
   y_train = numpy.array(Y_list, dtype=numpy.float32)

   x_train_t = x_train.transpose()
   x_train_mul = numpy.dot(x_train_t,x_train)   
   x_train_inv = inv(x_train_mul)   
   x_result_ma = numpy.dot(x_train_inv, x_train_t)   
   L_para = numpy.dot(x_result_ma, y_train)
   
   # print('LR:',L_para)
   
   
   L_min_comb = {}; L_min_comb2 = {}
   t_total = 0.0
   for idx in range(len(L_grp_order)):
      gid = L_grp_order[idx]
      va = L_para[idx]
      t_total += va
      L_min_comb[gid] = va

   for idx in range(len(L_grp_order)):
      gid = L_grp_order[idx]
      va = L_para[idx]
      #L_min_comb2[gid] = va/t_total*100  
      L_min_comb2[gid] = va/t_total 
     
   L_std_gene_new_dic = {}
   for gene in L_gene_list:
      t_va = 0.0
      for grp in L_min_comb:
         if L_min_comb[grp]>0:
            t_va += L_min_comb[grp]*p_mean_std[gene][grp][1]*p_mean_std[gene][grp][1]
      t_std = t_va**0.5
      L_std_gene_new_dic[gene] = t_std
   
   return L_min_comb2,L_min_comb, L_std_gene_new_dic


def get_cell_partition_Linear_Regression_Model_2_new(p_expression_RNAseq, p_mean_std, p_std_gene_old):
   #take absolute fraction value when computing variance (considering cell types with negative fractions)
   L_min_comb = {}

   L_grp_order = []
   for gene in p_mean_std:
      for grp in p_mean_std[gene]:
         L_grp_order.append(grp)
      break
   Y_list = []
   X_list = []
   L_gene_list = []
   for gene in p_expression_RNAseq:
      L_gene_list.append(gene)
   for gene in L_gene_list:
       t_gene_std = p_std_gene_old[gene] + 0.00000001
       Y_list.append(p_expression_RNAseq[gene]/t_gene_std)
       t_list = []
       for grp in   L_grp_order:
         t_list.append(p_mean_std[gene][grp][0]/t_gene_std)
       X_list.append(t_list)
   #print('----Length of Y:', len(Y_list))
   x_train = numpy.array(X_list, dtype=numpy.float32)
   y_train = numpy.array(Y_list, dtype=numpy.float32)

   x_train_t = x_train.transpose()
   x_train_mul = numpy.dot(x_train_t,x_train)   
   x_train_inv = inv(x_train_mul)   
   x_result_ma = numpy.dot(x_train_inv, x_train_t)   
   L_para = numpy.dot(x_result_ma, y_train)
   
   # print('LR:',L_para)
   
   
   L_min_comb = {}; L_min_comb2 = {}
   t_total = 0.0
   for idx in range(len(L_grp_order)):
      gid = L_grp_order[idx]
      va = L_para[idx]
      t_total += va
      L_min_comb[gid] = va

   for idx in range(len(L_grp_order)):
      gid = L_grp_order[idx]
      va = L_para[idx]
      #L_min_comb2[gid] = va/t_total*100  
      L_min_comb2[gid] = va/t_total 
     
   L_std_gene_new_dic = {}
   for gene in L_gene_list:
      t_va = 0.0
      for grp in L_min_comb:
         #if L_min_comb[grp]>0:
         t_va += abs(L_min_comb[grp])*p_mean_std[gene][grp][1]*p_mean_std[gene][grp][1]
      t_std = t_va**0.5
      L_std_gene_new_dic[gene] = t_std
   
   return L_min_comb2,L_min_comb, L_std_gene_new_dic


def get_cell_partition_Linear_Regression_Model_nnls(p_expression_RNAseq, p_mean_std, p_std_gene_old):
   L_min_comb = {}

   L_grp_order = []
   for gene in p_mean_std:
      for grp in p_mean_std[gene]:
         L_grp_order.append(grp)
      break
   Y_list = []
   X_list = []
   L_gene_list = []
   for gene in p_expression_RNAseq:
      L_gene_list.append(gene)
   for gene in L_gene_list:
       t_gene_std = p_std_gene_old[gene] + 0.00000001
       Y_list.append(p_expression_RNAseq[gene]/t_gene_std)
       t_list = []
       for grp in   L_grp_order:
         t_list.append(p_mean_std[gene][grp][0]/t_gene_std)
       X_list.append(t_list)
   x_train = numpy.array(X_list, dtype=numpy.float32)
   y_train = numpy.array(Y_list, dtype=numpy.float32)

   L_para, this_Res = nnls(x_train,y_train)

   
 #   paraTp = ''
 #   for item in L_para:
 #      paraTp += '\t'+str(item)
 #   print('---LR:',paraTp)
   
   
   L_min_comb = {}; L_min_comb2 = {}
   t_total = 0.0
   for idx in range(len(L_grp_order)):
       gid = L_grp_order[idx]
       va = L_para[idx]
       t_total += va
       L_min_comb[gid] = va

   for idx in range(len(L_grp_order)):
       gid = L_grp_order[idx]
       va = L_para[idx]
       #L_min_comb2[gid] = va/t_total*100  
       L_min_comb2[gid] = va/t_total 
     
   L_std_gene_new_dic = {}
   for gene in L_gene_list:
       t_va = 0.0
       for grp in L_min_comb:
         if L_min_comb[grp]>0:
             t_va += L_min_comb[grp]*p_mean_std[gene][grp][1]*p_mean_std[gene][grp][1]
       t_std = t_va**0.5
       L_std_gene_new_dic[gene] = t_std
   
   return L_min_comb2,L_min_comb, L_std_gene_new_dic


def Percentage_Fraction_nor(p_per, p_fra):
   L_percentage = {}; L_fraction = {}
   L_total = 0.0
   for item in p_per:
      item2 = p_per[item]
      if item2>0:
         L_total += item2
   for item in p_per:
      item2 = p_per[item]
      L_percentage[item] = max(0,item2/L_total)

   for item in p_fra:
      item2 = p_fra[item]
      L_fraction[item] = max(0,item2)
   return L_percentage, L_fraction

def MLMD(p_expression_RNAseq, p_mean_std, p_loop = 6):
   #Minimum Likelihood Model for Deconvolution
   #Fractions are non negative number
   L_percentage_all = {}; L_fractions_all = {}
   t_predic_percent_2 = []; t_fraction_2 = []
   for sp in p_expression_RNAseq:
      t_spot_sp = p_expression_RNAseq[sp]
      L_std_gene_old = {}
      for gene in p_mean_std:
          L_std_gene_old[gene] = 1.0
      for tloop in range(p_loop):
         t_predic_percent_2, t_fraction_2, t_std_gene_2 = get_cell_partition_Linear_Regression_Model_2_new(t_spot_sp, p_mean_std, L_std_gene_old)
         L_percentage_all[sp] = t_predic_percent_2
         L_fractions_all[sp] = t_fraction_2
         L_std_gene_old = t_std_gene_2
      L_per2, L_fra2 = Percentage_Fraction_nor( t_predic_percent_2, t_fraction_2)
      L_percentage_all[sp] = L_per2
      L_fractions_all[sp] = L_fra2
      
   return L_percentage_all, L_fractions_all


def LRM_Least_Square_Error(p_expression_RNAseq, p_mean_std):
   #Linear Regression Modle with Least Squre Error
   #Same with CIBERSORT, but not use the SVR algorithm
   L_percentage_all = {}; L_fractions_all = {}
   for sp in p_expression_RNAseq:
      t_spot_sp = p_expression_RNAseq[sp]
      L_std_gene_old = {}
      for gene in p_mean_std:
          L_std_gene_old[gene] = 1.0
      t_predic_percent_1, t_fraction_1, t_std_gene_1 = get_cell_partition_Linear_Regression_Model_2(t_spot_sp, p_mean_std, L_std_gene_old)
      L_percentage_all[sp] = t_predic_percent_1
      L_fractions_all[sp] = t_fraction_1
      
   return L_percentage_all, L_fractions_all



def MLMD_nns(p_expression_RNAseq, p_mean_std, p_loop = 6):
    #MLMD with non negative fraction predictions
    L_percentage_all = {}; L_fractions_all = {}
    #L_percentage_allB = {}; L_fractions_allB = {}
    for sp in p_expression_RNAseq:
      t_spot_sp = p_expression_RNAseq[sp]
      L_std_gene_old = {}
      for gene in p_mean_std:
          L_std_gene_old[gene] = 1.0
      for t_loop in range(p_loop):
          t_predic_percent_2, t_fraction_2, t_std_gene_2 = get_cell_partition_Linear_Regression_Model_nnls(t_spot_sp, p_mean_std, L_std_gene_old)
          L_percentage_all[sp] = t_predic_percent_2
          L_fractions_all[sp] = t_fraction_2
          L_std_gene_old = t_std_gene_2

      
def Save_results(p_results, p_fn_out):
   Cell_type = []
   #L_sample_list = []; L_ctp_sp_per = {}
   
   fout = open(p_fn_out,'w')
   
   count = 0
   for sp in  p_results:
      count += 1
      #L_sample_list.append(sp)
      if count == 1:
         for ctp in p_results[sp]:
            Cell_type.append(ctp)
            #L_ctp_sp_per[ctp] = []
         Cell_type.sort()
         if (p_st_using.upper())[0] == 'Y':
            print('+++ cell types:', Cell_type)
   
         #output  = 'Sample\Cell_type'
         output  = 'sample/cell_type'
         for item in Cell_type:
            output += '\t'+item
         fout.write(output)
         fout.write('\n')
               
      output = sp
      t_prediction = p_results[sp]
      for ctp in Cell_type:
          output += '\t'+str(t_prediction[ctp])
          #L_ctp_sp_per[sp].append(t_prediction[ctp])

      fout.write(output)
      fout.write('\n')

   fout.close()
   #return L_sample_list, L_ctp_sp_per

     
def ReDeconv(p_sig_matrix, p_bulk_exp, p_fn_save):
   #Our new model for cell type deconvolution
   #read mean and std of top signature genes in each cell type
   L_mean_std, L_cell_type = read_sig_gene_mean_std(p_sig_matrix)  
    
   #read expression data for the mixed samples
   L_samples, L_mean_std_new = read_bulk_RNA_seq_column_for_sample(p_bulk_exp, L_mean_std)
   
   #Find cell type percentages and fractions in all 
   L_predicted_percentages, L_predicted_fractions = MLMD(L_samples, L_mean_std_new,6)

   
   #save cell type perventage predictions
   Save_results(L_predicted_percentages,p_fn_save)


   
def LRM_LSE(p_sig_matrix, p_bulk_exp, p_fn_save):
   #Linear Regression Modle with Least Squre Error
   
   #read mean and std of top signature genes in each cell type
   L_mean_std, L_cell_type = read_sig_gene_mean_std(p_sig_matrix)  
    
   #read expression data for the mixed samples
   L_samples, L_mean_std_new = read_bulk_RNA_seq_column_for_sample(p_bulk_exp, L_mean_std)
   
   #Find cell type percentages and fractions in all 
   L_predicted_percentages, L_predicted_fractions = LRM_Least_Square_Error(L_samples, L_mean_std_new)
   
   #save cell type perventage predictions
   Save_results(L_predicted_percentages,p_fn_save)



from setuptools import setup, find_packages

setup(
    name="redeconv",
    version="1.2.1",
    packages=find_packages(),
    install_requires=[
        "matplotlib==3.7.2",
        "numpy==1.24.4",
        "pandas==2.0.3",
        "scipy==1.10.1",
        "seaborn==0.12.2",
    ],
    python_requires=">=3.7",
    author="Songjian Lu",
    author_email="songjian.lu@stjude.org",
    description="One wildly used method for scRNA-seq data normalization is to make the total count, called the transcriptome size, of all genes in each cell to be the same, such as counts per million (CPM) or count per 10 thousand (CP10K)",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jyyulab/redeconv",
)
